"""Data_reader module containing corpus construction utilities for PVDBOW (skipgram) models.

This module describes the classes which handle graph corpi and datasets
which can be loaded into PyTorch dataloaders. 

"""

import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from collections import defaultdict, Counter
from random import shuffle
from tqdm import tqdm

from .utils import get_files

np.random.seed(27)

#######################################################################################################
#######################################################################################################
# Harddrive version
#######################################################################################################
#######################################################################################################
class PVDBOWCorpus(Dataset):
	"""Class which represents the target-context dataset created
	over the graph documents for PVDBOW models. In this version 
	the __getitem__ function loads individual target-context pairs
	from the hard-drive. As a result, it is quick to set up and 
	memory efficient but may perform slower in training time.

	Parameters
	----------
	corpus_dir : str
		path to folder with graph document files created in decomposition stage
	extension : str
		extension of the graph document files from which the corpus should be built
	max_files : int (default=0)
		the maximum number of files to include. Useful for debugging or other 
		artificial scenarios. The default of 0 includes all files with matching
		extension
	window_size : int (default=1)
		The number of context substructure patterns to be considered for every target. 
		This needs to be greater than 0.

	Returns
	-------
	self : PVDBOWCorpus 
		A corpus dataset that can be used with the skipgram with negative 
		sampling model to learn graph-level embeddings.
	"""
	NEGATIVE_TABLE_SIZE=1e8

	def __init__(self, corpus_dir=None, extension=".wld2", max_files=0, min_count=0):
		assert corpus_dir != None, "Please specify the path where the graph files are"
		self.corpus_dir = corpus_dir
		self.extension = extension
		self.graph_index = 0
		self.subgraph_index = 0
		self.epoch_flag = 0
		self.max_files = max_files
		self.graph_ids_for_batch_traversal = []
		self.min_count = min_count

		self.negatives = []
		self.discards = []
		self.negpos = 0

		# Methods
		self.scan_and_load_corpus()
		self.initTableDiscards()
		self.initTableNegatives()

	def scan_and_load_corpus(self):
		"""Gets the list of graph file paths, gives them number ids in a map and calls 
		scan_corpus also makes available a list of shuffled graph_ids for batch
		"""
		print("#... Scanning and loading corpus from %s" % (self.corpus_dir))
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)
		self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)}
		self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}

		# Scan the corpus
		# This creates the alphabet and vocabulary of subgraphs
		subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

		self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
		shuffle(self.graph_ids_for_batch_traversal)

	def scan_corpus(self, min_count):
		"""Maps the graph files to a subgraph alphabet from which we create new_ids for the subgraphs
		which in turn get used by the skipgram architectures

		Parameters
		----------
		min_count : int
			The minimum number of times a subgraph pattern should appear across the graphs in 
			order to be considered part of the vocabulary.

		Returns
		-------
		(Optional) self._subgraph_to_id_map : dict
			dictionary of substructure pattern to int id map

		"""

		# Get all the subgraph names (the centers, ie without context around itself) ie first item of 
		# each line in the grapdoc files
		subgraphs = []
		for fname in self.graph_fname_list:
			subgraphs.extend([l.split()[0] for l in open(fname).readlines()])
		subgraphs.append('UNK') # end flag

		# Frequency map of subgraph types
		subgraph_to_freq_map = Counter(subgraphs)
		del subgraphs

		# Remove infrequent graph patterns if user specified a min_count
		if min_count:
			subgraph_to_freq_map = dict(subgraph_to_freq_map)
			for key in list(subgraph_to_freq_map.keys()):
				if subgraph_to_freq_map[key] < min_count:
					subgraph_to_freq_map.pop(key, None)

		# Also give each of the subgraph labels new int ids
		subgraph_to_id_map = {sg:i for i, sg, in enumerate(subgraph_to_freq_map.keys())}

		self._subgraph_to_freq_map = subgraph_to_freq_map
		self._subgraph_to_id_map = subgraph_to_id_map
		self._id_to_subgraph_map = {v:k  for k,v in subgraph_to_id_map.items()}
		self._subgraphcount = sum(subgraph_to_freq_map.values()) # token count, the number of subgraphs in all graphs


		self.num_graphs = len(self.graph_fname_list) # number of targets
		self.num_subgraphs = len(subgraph_to_id_map) # size of vocabulary

		# This is a list sorted by id, which contains the frequency of a subgraph appearing as a list
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])

		return self._subgraph_to_id_map

	def add_file(self, full_graph_path):
		"""This method is used to add new graphs into the corpus for 
		inductive learning of new unseen graphs

		Parameters
		----------
		full_graph_path : str
			path to graph document to be part of the new corpus

		Returns
		-------
		None
			New graph and its substructure patterns is made part of the corpus
		"""

		# Retrieve the graphs files and assign them internal ids for this method
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)

		if full_graph_path in self.graph_fname_list:
			# if the graph is already in the corpus we may ignore it
			print("The graph %s is already in the corpus" % (full_graph_path))
			return
		else:
			self.graph_fname_list.append(full_graph_path) # add the new file
			self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)} # add to the end
			self._graph_name_to_id_map[full_graph_path] = self.num_graphs
			self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}
				
			# Scan the corpus (ie explain in more detail in a sec)
			# This creates an "alphabet" and "vocabulary" of subgraphs
			subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

			self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
			shuffle(self.graph_ids_for_batch_traversal)

			self.initTableDiscards()
			self.initTableNegatives()

	# should run this as part of init
	def initTableDiscards(self):
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])
		self.subgraph_id_freq_map = {}
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map[i] = self.subgraph_id_freq_map_as_list[i]

		t = 1e-4
		f = np.array(list(self.subgraph_id_freq_map.values())) / self._subgraphcount
		self.discards = np.sqrt(t/f)+(t/f)

	# Needs translations but line complete
	def initTableNegatives(self):
		pow_frequency = np.array(list(self.subgraph_id_freq_map.values()))**0.5
		words_pow = sum(pow_frequency)
		ratio = pow_frequency / words_pow
		count = np.round(ratio * PVDBOWCorpus.NEGATIVE_TABLE_SIZE)
		for sg_id, c in enumerate(count):
			self.negatives += [sg_id]*int(c)
		self.negatives = np.array(self.negatives)
		np.random.shuffle(self.negatives)
		
	def getNegatives(self, target, size):
		r"""Given target find a `size` number of negative samples by index

		Parameters
		----------
		target : int
			internal int id of the subgraph pattern
		size : int
			number of negative samples to find

		Returns
		-------
		response : [int]
			list of negative samples by internal int id

		"""

		response = self.negatives[self.negpos:self.negpos + size]
		self.negpos = (self.negpos + size) % len(self.negatives)
		while target in response: # check equality with target
			for i in np.where(response == target):
				response[i] = self.negatives[np.random.randint(0,len(self.negatives))]
		if len(response) != size:
			return np.concatenate((response, self.negatives[0:self.negpos]))
		return response

	def __len__(self):
		"""Return the number of total number of subgraphs
		
		"""
		# print("The number of graph-subgraph pairs considering batchsize in the dataset is")
		return self._subgraphcount

	def __getitem__(self, idx):
		""" Get a single target-context observation from the dataset.

		This version loads each individual item from the hard drive

		"""
		target_graph_ids = []
		context_subgraph_ids = []

		# Extract a random graph and read its contents
		graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]] # pull out a random graph (its filename here)
		graph_contents = open(graph_name).readlines()
		
		# tldr: random graph traverser 
		# If we've looked at all the subgraphs we go to the next graph
		# We set the epoch flag true if we've also gone through n graphs (in a n graph dataset)
		# we then grab the next graph file whether that be the next graph in the shuffled list
		# or the first graph in a reshuffled list of graph files
		while self.subgraph_index >= len(graph_contents):
			self.subgraph_index = 0
			self.graph_index += 1
			if self.graph_index == len(self.graph_fname_list):
				self.graph_index = 0
				np.random.shuffle(self.graph_ids_for_batch_traversal)
				self.epoch_flag = True
			graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
			graph_contents = open(graph_name).readlines()

		# Given that we haven't gotten enough graphs for our batch
		# We traverse the file at graph_name and graph the center as the (context subgraph)
		# which is a bit counter-intuitive but we consider the centers as the "context" of the
		# graph as a whole.
		# while len(context_subgraph_ids) < batch_size:
		while len(context_subgraph_ids) < 1:
			line_id = self.subgraph_index
			context_subgraph = graph_contents[line_id].split()[0] # first item on the line
			target_graph = graph_name

			if context_subgraph in self._subgraph_to_id_map:
				context_subgraph_ids.append(self._subgraph_to_id_map[context_subgraph]) # add the ids of the subgraph into the context
				target_graph_ids.append(self._graph_name_to_id_map[target_graph]) # add the id of the graph for the target

			# context_subgraph_ids.append(self._subgraph_to_id_map[context_subgraph]) # add the ids of the subgraph into the context
			# target_graph_ids.append(self._graph_name_to_id_map[target_graph]) # add the id of the graph for the target

			# move on to the next subgraph
			self.subgraph_index += 1

			# if we've reached the end of the graph file (ie exhasuted all the subgraphs in it)
			# we move onto the next graph as above.
			while self.subgraph_index == len(graph_contents):
				self.subgraph_index = 0
				self.graph_index +=1 
				if self.graph_index == len(self.graph_fname_list):
					self.graph_index = 0
					np.random.shuffle(self.graph_ids_for_batch_traversal)
					self.epoch_flag = True

				graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
				graph_contents = open(graph_name).readlines()
		
		# Once we've built 'batch_size' number worth of targets and contexts
		# we zip them, shuffle them, and unzip the pairs in shuffled order (keeping the pairing)
		target_context_pairs = list(zip(target_graph_ids, context_subgraph_ids))
		shuffle(target_context_pairs)
		target_graph_ids, context_subgraph_ids = list(zip(*target_context_pairs))

		negatives_per_context = [self.getNegatives(x,10) for x in context_subgraph_ids]
		target_context_negatives = [(target, context, negatives) for (target, context, negatives) in zip(target_graph_ids, context_subgraph_ids, negatives_per_context)]
		return target_context_negatives

	@staticmethod
	def collate(batches):
		all_targets = [target for batch in batches for target, _, _ in batch if len(batch)>0]
		all_contexts = [context for batch in batches for _, context, _ in batch if len(batch)>0]
		all_neg_contexts = [neg_context for batch in batches for _, _, neg_context in batch if len(batch)>0]

		return torch.LongTensor(all_targets), torch.LongTensor(all_contexts), torch.LongTensor(all_neg_contexts)

#######################################################################################################
#######################################################################################################
# In memory version of skipgram datareader
#######################################################################################################
#######################################################################################################

class PVDBOWInMemoryCorpus(Dataset):
	"""Class which represents the target-context dataset created
	over the graph documents for PVDBOW models. This version keeps 
	the entire corpus with negatives in memory which requires a 
	larger initial creation time but has a much quicker __getitem__ 
	computation.

	Parameters
	----------
	corpus_dir : str
		path to folder with graph document files created in decomposition stage
	extension : str
		extension of the graph document files from which the corpus should be built
	max_files : int (default=0)
		the maximum number of files to include. Useful for debugging or other 
		artificial scenarios. The default of 0 includes all files with matching
		extension
	window_size : int (default=1)
		The number of context substructure patterns to be considered for every target. 
		This needs to be greater than 0.

	Returns
	-------
	self : PVDBOWInMemoryCorpus 
		A corpus dataset that can be used with the skipgram with negative 
		sampling model to learn graph-level embeddings.
	"""
	NEGATIVE_TABLE_SIZE=1e8

	def __init__(self, corpus_dir=None, extension=".wld2", max_files=0, min_count=0):
		assert corpus_dir != None, "Please specify the path where the graph files are"
		self.corpus_dir = corpus_dir
		self.extension = extension
		self.graph_index = 0
		self.subgraph_index = 0
		self.epoch_flag = 0
		self.max_files = max_files
		self.graph_ids_for_batch_traversal = []
		self.min_count = min_count

		self.negatives = []
		self.discards = []
		self.negpos = 0

		# Methods
		self.scan_and_load_corpus()
		self.initTableDiscards()
		self.initTableNegatives()
		self.pre_load_corpus()

	def scan_and_load_corpus(self):
		"""Gets the list of graph file paths, gives them number ids in a map and calls 
		scan_corpus also makes available a list of shuffled graph_ids

		"""

		print("#... Scanning and loading corpus from %s" % (self.corpus_dir))
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)
		self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)}
		self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}

		# Scan the corpus
		# This creates the alphabet and vocabulary of subgraphs
		subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

		self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
		shuffle(self.graph_ids_for_batch_traversal)

	def scan_corpus(self, min_count):
		"""Maps the graph files to a subgraph alphabet from which we create new_ids for the subgraphs
		which in turn get used by the skipgram architectures

		Parameters
		----------
		min_count : int
			The minimum number of times a subgraph pattern should appear across the graphs in 
			order to be considered part of the vocabulary.

		Returns
		-------
		(Optional) self._subgraph_to_id_map : dict
			dictionary of substructure pattern to int id map

		"""

		# Get all the subgraph names (the centers, ie without context around itself) ie first item of 
		# each line in the grapdoc files
		subgraphs = []
		for fname in self.graph_fname_list:
			subgraphs.extend([l.split()[0] for l in open(fname).readlines()])
		subgraphs.append('UNK') # end flag

		# Frequency map of subgraph types
		subgraph_to_freq_map = Counter(subgraphs)
		del subgraphs

		# Remove infrequent graph patterns if user specified a min_count
		if min_count:
			subgraph_to_freq_map = dict(subgraph_to_freq_map)
			for key in list(subgraph_to_freq_map.keys()):
				if subgraph_to_freq_map[key] < min_count:
					subgraph_to_freq_map.pop(key, None)

		# Also give each of the subgraph labels new int ids
		subgraph_to_id_map = {sg:i for i, sg, in enumerate(subgraph_to_freq_map.keys())}

		self._subgraph_to_freq_map = subgraph_to_freq_map
		self._subgraph_to_id_map = subgraph_to_id_map
		self._id_to_subgraph_map = {v:k  for k,v in subgraph_to_id_map.items()}
		self._subgraphcount = sum(subgraph_to_freq_map.values()) # token count, the number of subgraphs in all graphs


		self.num_graphs = len(self.graph_fname_list) # number of targets
		self.num_subgraphs = len(subgraph_to_id_map) # size of vocabulary

		# This is a list sorted by id, which contains the frequency of a subgraph appearing as a list
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])

		return self._subgraph_to_id_map

	def add_file(self, full_graph_path):
		"""This method is used to add new graphs into the corpus for 
		inductive learning of new unseen graphs

		Parameters
		----------
		full_graph_path : str
			path to graph document to be part of the new corpus

		Returns
		-------
		None
			New graph and its substructure patterns is made part of the corpus
		"""

		# Retrieve the graphs files and assign them internal ids for this method
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)

		if full_graph_path in self.graph_fname_list:
			# if the graph is already in the corpus we may ignore it
			print("The graph %s is already in the corpus" % (full_graph_path))
			return
		else:
			self.graph_fname_list.append(full_graph_path) # add the new file
			self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)} # add to the end
			self._graph_name_to_id_map[full_graph_path] = self.num_graphs
			self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}
				
			# Scan the corpus (ie explain in more detail in a sec)
			# This creates an "alphabet" and "vocabulary" of subgraphs
			subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

			self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
			shuffle(self.graph_ids_for_batch_traversal)

			self.initTableDiscards()
			self.initTableNegatives()

	# should run this as part of init
	def initTableDiscards(self):
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])
		self.subgraph_id_freq_map = {}
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map[i] = self.subgraph_id_freq_map_as_list[i]

		t = 1e-4
		f = np.array(list(self.subgraph_id_freq_map.values())) / self._subgraphcount
		self.discards = np.sqrt(t/f)+(t/f)

	# Needs translations but line complete
	def initTableNegatives(self):
		pow_frequency = np.array(list(self.subgraph_id_freq_map.values()))**0.5
		words_pow = sum(pow_frequency)
		ratio = pow_frequency / words_pow
		count = np.round(ratio * PVDBOWInMemoryCorpus.NEGATIVE_TABLE_SIZE)
		for sg_id, c in enumerate(count):
			self.negatives += [sg_id]*int(c)
		self.negatives = np.array(self.negatives)
		np.random.shuffle(self.negatives)

	def getNegatives(self, target, size):
		r"""Given target find a `size` number of negative samples by index

		Parameters
		----------
		target : int
			internal int id of the subgraph pattern
		size : int
			number of negative samples to find

		Returns
		-------
		response : [int]
			list of negative samples by internal int id

		"""

		response = self.negatives[self.negpos:self.negpos + size]
		self.negpos = (self.negpos + size) % len(self.negatives)
		while target in response: # check equality with target
			for i in np.where(response == target):
				response[i] = self.negatives[np.random.randint(0,len(self.negatives))]
		if len(response) != size:
			return np.concatenate((response, self.negatives[0:self.negpos]))
		return response


	def pre_load_corpus(self):
		"""Constructs and loads an entire context-pair dataset into memory
		
		"""

		print("#... Generating dataset in memory for quick dataloader access")
		# Get a single item of data from the dataset.
		self.context_pair_dataset = []

		for c in tqdm(range(self._subgraphcount)):
			target_graph_ids = []
			context_subgraph_ids = []

			# Extract a random graph and read its contents
			graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]] # pull out a random graph (its filename here)
			graph_contents = open(graph_name).readlines()
			
			# tldr: random graph traverser 
			# If we've looked at all the subgraphs we go to the next graph
			# We set the epoch flag true if we've also gone through n graphs (in a n graph dataset)
			# we then grab the next graph file whether that be the next graph in the shuffled list
			# or the first graph in a reshuffled list of graph files
			while self.subgraph_index >= len(graph_contents):
				self.subgraph_index = 0
				self.graph_index += 1
				if self.graph_index == len(self.graph_fname_list):
					self.graph_index = 0
					np.random.shuffle(self.graph_ids_for_batch_traversal)
					self.epoch_flag = True
				graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
				graph_contents = open(graph_name).readlines()

			# Given that we haven't gotten enough graphs for our batch
			# We traverse the file at graph_name and graph the center as the (context subgraph)
			# which is a bit counter-intuitive but we consider the centers as the "context" of the
			# graph as a whole.
			# while len(context_subgraph_ids) < batch_size:
			# while len(context_subgraph_ids) < 1:
			line_id = self.subgraph_index
			context_subgraph = graph_contents[line_id].split()[0] # first item on the line
			target_graph = graph_name

			if context_subgraph in self._subgraph_to_id_map:
				context_subgraph_ids.append(self._subgraph_to_id_map[context_subgraph]) # add the ids of the subgraph into the context
				target_graph_ids.append(self._graph_name_to_id_map[target_graph]) # add the id of the graph for the target

			# context_subgraph_ids.append(self._subgraph_to_id_map[context_subgraph]) # add the ids of the subgraph into the context
			# target_graph_ids.append(self._graph_name_to_id_map[target_graph]) # add the id of the graph for the target

			# move on to the next subgraph
			self.subgraph_index += 1

			# if we've reached the end of the graph file (ie exhasuted all the subgraphs in it)
			# we move onto the next graph as above.
			while self.subgraph_index == len(graph_contents):
				self.subgraph_index = 0
				self.graph_index +=1 
				if self.graph_index == len(self.graph_fname_list):
					self.graph_index = 0
					np.random.shuffle(self.graph_ids_for_batch_traversal)
					self.epoch_flag = True

				graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
				graph_contents = open(graph_name).readlines()
			
			# Once we've built 'batch_size' number worth of targets and contexts
			# we zip them, shuffle them, and unzip the pairs in shuffled order (keeping the pairing)
			target_context_pairs = list(zip(target_graph_ids, context_subgraph_ids))
			shuffle(target_context_pairs)
			target_graph_ids, context_subgraph_ids = list(zip(*target_context_pairs))

			negatives_per_context = [self.getNegatives(x,10) for x in context_subgraph_ids]
			target_context_negatives = [(target, context, negatives) for (target, context, negatives) in zip(target_graph_ids, context_subgraph_ids, negatives_per_context)]
			self.context_pair_dataset.append(target_context_negatives)
		self.epoch_flag = False

	def __len__(self):
		"""Return the number of total number of subgraphs
		
		"""
		return self._subgraphcount

	def __getitem__(self, idx):
		""" Get a single target-context observation from the dataset in memory.

		"""
		return self.context_pair_dataset[idx]

	@staticmethod
	def collate(batches):
		all_targets = [target for batch in batches for target, _, _ in batch if len(batch)>0]
		all_contexts = [context for batch in batches for _, context, _ in batch if len(batch)>0]
		all_neg_contexts = [neg_context for batch in batches for _, _, neg_context in batch if len(batch)>0]

		return torch.LongTensor(all_targets), torch.LongTensor(all_contexts), torch.LongTensor(all_neg_contexts)



if __name__ == '__main__':
	corpus_dir = "../data/dortmund_gexf/MUTAG"

	# Normal Harddrive corpus
	corpus = PVDBOWCorpus(corpus_dir, extension=".wld2", max_files=60)
	# additional_file = "/home/morio/workspace/geo2dr/geometric2dr/file_handling/dortmund_gexf/MUTAG/99.gexf.wld2"
	# corpus.add_file(additional_file)
	dataloader = DataLoader(corpus, batch_size=4, shuffle=False, num_workers=0, collate_fn=corpus.collate)

	# In memory corpus
	in_mem_corpus = PVDBOWInMemoryCorpus(corpus_dir, extension=".wld2", max_files=60)
	mem_dataloader = DataLoader(in_mem_corpus, batch_size=4, shuffle=False, num_workers=0, collate_fn=in_mem_corpus.collate)


"""
A trainer class which faciliates training of the embedding methods by the set hyperparameters.

Author: Paul Scherer 2020
"""
import numpy as np
import torch
import torch.optim as optim
import torch.nn as nn
from torch.utils.data import DataLoader
from tqdm import tqdm

# Internal 
from .pvdm_data_reader import PVDMCorpus
from .pvdm import PVDM
from .utils import save_graph_embeddings

# For testing
from .classify import perform_classification, cross_val_accuracy

class PVDM_Trainer(object):
	"""Handles corpus construction, CBOW initialization and training.

	Paramaters
	----------
	corpus_dir : str
		path to directory containing graph files
	extension : str
		extension used in graph documents produced after decomposition stage
	max_files : int
		the maximum number of graph files to consider, default of 0 uses all files
	window_size : int
		the number of cooccuring context subgraph patterns to use
	output_fh : str
		the path to the file where embeddings should be saved
	emb_dimension : int (default=128)
		the desired dimension of the embeddings
	batch_size : int (default=32)
		the desired batch size
	epochs : int (default=100)
		the desired number of epochs for which the network should be trained
	initial_lr : float (default=1e-3)
		the initial learning rate
	min_count : int (default=1)
		the minimum number of times a pattern should occur across the dataset to 
		be considered part of the substructure pattern vocabulary

	Returns
	-------
	self : PVDM_Trainer
		A PVDM_Trainer instance
	"""

	def __init__(self, corpus_dir, extension, max_files, window_size, output_fh, emb_dimension=128, batch_size=32, epochs=100, initial_lr=1e-3, min_count=1):
		self.corpus = PVDMCorpus(corpus_dir, extension, max_files, min_count, window_size)
		self.dataloader = DataLoader(self.corpus, batch_size, shuffle=False, num_workers=0, collate_fn = self.corpus.collate)
		
		self.corpus_dir = corpus_dir
		self.extension = extension
		self.max_files = max_files
		self.output_fh = output_fh
		self.emb_dimension = emb_dimension
		self.batch_size = batch_size
		self.epochs = epochs
		self.initial_lr = initial_lr
		self.min_count = min_count
		self.window_size = window_size

		self.num_targets = self.corpus.num_graphs
		self.vocab_size = self.corpus.num_subgraphs

		self.pvdm = PVDM(self.num_targets, self.vocab_size, self.emb_dimension)

		if torch.cuda.is_available():
			self.device = torch.device("cuda")
			self.pvdm.cuda()
		else:
			self.device = torch.device("cpu")

	def train(self):
		"""Train the network with the settings used to initialise the PVDM_Trainer
		
		"""

		for epoch in range(self.epochs):
			print("### Epoch: " + str(epoch))
			criterion = nn.CrossEntropyLoss()
			optimizer = optim.Adagrad(self.pvdm.parameters(), lr=self.initial_lr, lr_decay=0.00001)

			running_loss = 0.0
			for i, sample_batched in enumerate(tqdm(self.dataloader)):

				if len(sample_batched[0]) > 1:
					pos_target_graph = sample_batched[0].to(self.device)
					pos_target_subgraph = sample_batched[1].to(self.device)
					pos_contexts_for_subgraph_target = sample_batched[2].to(self.device)
					pos_negatives = sample_batched[3].to(self.device)

					optimizer.zero_grad()
					loss = self.pvdm.forward(pos_target_graph, pos_target_subgraph, pos_contexts_for_subgraph_target, pos_negatives)
					loss.backward()

					# log_probs = self.pvdm(pos_target_graph, pos_target_subgraph, pos_contexts_for_subgraph_target, pos_negatives)
					# print(log_probs)
					# print("Most Likely Class %s " % ([np.argmax(x.detach().numpy()) for x in log_probs] ) ) 
					# print("True Class %s " % (pos_target_subgraph.detach().numpy()))
					# loss = criterion(log_probs, torch.tensor(pos_target_subgraph, dtype=torch.long))

					optimizer.step()

					running_loss = running_loss * 0.9 + loss.item() * 0.1
			print(" Loss: " + str(running_loss))

		final_embeddings = self.pvdm.target_embeddings.weight.cpu().data.numpy()
		save_graph_embeddings(self.corpus, final_embeddings, self.output_fh)

# Some test code
if __name__ == '__main__':

	corpus_dir = "../data/dortmund_gexf/MUTAG" # A needed parameter
	extension = ".awe_8_nodes"
	output_file = "PVDMEmbeddings.json" # A needed parameter
	emb_dimension = 64 # A needed parameter
	batch_size = 32 # A needed parameter
	epochs = 250 # A needed parameter
	initial_lr = 0.00001 # A needed parameter
	min_count= 0 # A needed parameter
	window_size = 5

	trainer = PVDM_Trainer(corpus_dir=corpus_dir, extension=extension, max_files=0, window_size=window_size,
					output_fh=output_file, emb_dimension=emb_dimension, batch_size=batch_size,
					epochs=epochs, initial_lr=initial_lr, min_count=min_count)
	trainer.train()

	final_embeddings = trainer.pvdm.give_target_embeddings()
	graph_files = trainer.corpus.graph_fname_list
	class_labels_fname = "../data/MUTAG.Labels"
	embedding_fname = trainer.output_fh
	classify_scores = cross_val_accuracy(corpus_dir, trainer.corpus.extension, embedding_fname, class_labels_fname)
	mean_acc, std_dev = classify_scores
	print("Mean accuracy using 10 cross fold accuracy: %s with std %s" % (mean_acc, std_dev))




"""Shortest path based decomposition algorithm to create graph documents. 
Inspired and adapted from Yanardag and Vishwanathan "Deep Graph Kernels" [2]_.

"""

# Copyright (c) 2016 Pinar Yanardag
#               2019 Paul Scherer

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Standard libraries
import sys
import os
import glob
import pickle
import itertools
import random
from collections import defaultdict
from time import time

# 3rd party
import numpy as np
import networkx as nx

# Internal
from .utils import get_files

# Random seeds from Yanardag et al.
random.seed(314124)
np.random.seed(2312312)

def load_graph(file_handle):
	""" Load the gexf format file as a nx_graph and an adjacency matrix.

	Parameters
	----------
	file_handle : str
	    path to gexf file

	Returns
	-------
	graph : networkx graph
	    Networkx graph instance of input gexf file
	adj_matrix : numpy ndarray
	    The adjacency matrix of the graph

	"""
	graph = nx.read_gexf(file_handle)
	adj_matrix = nx.to_numpy_matrix(graph)
	return graph, adj_matrix

def save_sp_doc(gexf_fh, gidx, coocurrence_corpus):
	"""Saves a shortest path graph doc with the extension .spp
	
	Parameters
	----------
	gexf_fh : str
		sdf
	gidx : int
		int id of the gexf (note: deprecated and will 
		be removed in next version)
	cooccurrence_corpus : list
		list of lists containing cooccuring shortest paths 
		(ie those starting from the same starting node)

	Returns
	-------
	None : None
		Saves the induced shortest path patterns into a graph document for the graph specified
		in the `gexf_fh` 

	"""

	open_fname = gexf_fh + ".spp"
	# if os.path.isfile(open_fname):
	# 	return
	with open(open_fname,'w') as fh:
		for spp_neighbourhood in coocurrence_corpus:
			sentence = str.join(" ", map(str, spp_neighbourhood))
			print (sentence, file=fh)

def sp_corpus(corpus_dir):
	"""Function which induces the shortest path patterns over the graphs inside a given directory.

	The main use case for this script is for the user to supply a path to the directory
	containing the .gexf graph files of a dataset. The decomposition function will produce
	a .spp file for each .gexf file which contains the shortest path patterns of a graph
	given the vocabulary of patterns across the dataset.

	Parameters
	----------
	corpus_dir : str
		path to directory containing graph files of dataset in .gexf format

	Returns
	-------
	corpus : list of lists of str
	    a list of lists, with each inner list containing all the shortest path patterns in one graph of the dataset
	vocabulary : list
	    a set of the unique shortest path pattern ids
	prob_map : dict
	    a map {gidx: {path: normalized_prob}} of normalized probabilities of a shortest path pattern appearing in a graph based on counts made in generation
	num_graphs : int
	    the number of graphs in the dataset
	graph_map : dict
	    a map {gidx: {path: count}} of the number of times a certain shortest path pattern appeared in a graph for each graph gidx in the dataset
	
	"""
	graph_files = get_files(corpus_dir, extension=".gexf")
	vocabulary = set() 
	count_map = {}
	graph_map = {}
	corpus = []

	for gexf_fh in graph_files:
		open_fname = gexf_fh + ".spp"
		if os.path.exists(open_fname):
			continue
		
		gidx = int((os.path.basename(gexf_fh)).replace(".gexf", ""))
		graph, am = load_graph(gexf_fh)

		count_map[gidx] = {}
		label_map = [graph.nodes[nidx]["Label"] for nidx in sorted(list(graph.nodes()))]
		coocurrence_corpus = []

		G = graph
		all_shortest_paths = nx.all_pairs_shortest_path(G)
		tmp_corpus = []
		
		# For each source and all the path endpoints add the label
		# We (Yanardag and subsequently Burnaev) consider paths to 
		# "cooccur" if they share the same source node
		for source, sink_map in list(all_shortest_paths):
			paths_cooccurrence = []
			for sink, path in sink_map.items():
				sp_length=len(path)-1
				label = "_".join(map(str, sorted([label_map[int(source)-1], label_map[int(sink)-1]]))) + "_" + str(sp_length)
				tmp_corpus.append(label)
				paths_cooccurrence.append(label)
				count_map[gidx][label] = count_map[gidx].get(label,0) + 1
				vocabulary.add(label)
			coocurrence_corpus.append(paths_cooccurrence)
		
		# Record and save the information
		save_sp_doc(gexf_fh, gidx, coocurrence_corpus)
		corpus.append(tmp_corpus)

	graph_map = count_map
	# When we are using a straight up MLE kernel we use the graph_map with the
	# counts instead
	prob_map = {gidx: {path: count/float(sum(paths.values())) \
		for path, count in paths.items()} for gidx, paths in count_map.items()}

	num_graphs = len(count_map)
	return corpus, vocabulary, prob_map, num_graphs, graph_map

if __name__ == '__main__':
	corpus_dir = "/home/morio/workspace/geo2dr/geometric2dr/data/dortmund_gexf/MUTAG/"
	corpus, vocabulary, prob_map, num_graphs, graph_map = sp_corpus(corpus_dir)


"""
A development trainer mainly for testing out functions when developing
"""

import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm

# internal
from pvdbow_data_reader import PVDBOWCorpus, PVDBOWInMemoryCorpus
from skipgram import Skipgram
from utils import get_class_labels, get_class_labels_tuples, save_graph_embeddings
from classify import perform_classification, cross_val_accuracy

corpus_dir = "../data/dortmund_gexf/MUTAG" # A needed parameter
corpus = PVDBOWCorpus(corpus_dir, extension=".wld2", max_files=0, min_count=0)
dataloader = DataLoader(corpus, batch_size=256, shuffle=False, num_workers=0, collate_fn=corpus.collate)

output_file = "Embeddings.testfile" # A needed parameter
output_file_name = output_file
num_targets = corpus.num_graphs
vocab_size = corpus.num_subgraphs
emb_dimension = 100 # A needed parameter
batch_size = 256 # A needed parameter
epochs = 100 # A needed parameter
initial_lr = 0.001 # A needed parameter
skipgram = Skipgram(num_targets, vocab_size, emb_dimension)

if torch.cuda.is_available():
	device = torch.device("cuda")
	skipgram.cuda()
else:
	device = torch.device("cpu")

for epoch in range(epochs):
	print("### Epoch: " + str(epoch))
	optimizer = optim.SparseAdam(skipgram.parameters(), lr=initial_lr)
	scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, len(dataloader))

	running_loss = 0.0
	for i, sample_batched in enumerate(tqdm(dataloader)):
	# for i, sample_batched in enumerate(dataloader):

		if len(sample_batched[0]) > 1:
			pos_target = sample_batched[0].to(device)
			pos_context = sample_batched[1].to(device)
			neg_context = sample_batched[2].to(device)

			optimizer.zero_grad()
			loss = skipgram.forward(pos_target, pos_context, neg_context) # the loss is integrated into the forward function
			loss.backward()
			optimizer.step()
			scheduler.step()

			running_loss = running_loss * 0.9 + loss.item() * 0.1
	print(" Loss: " + str(running_loss))

# skipgram.save_embedding(corpus._id_to_graph_name_map, output_file_name)

final_embeddings = skipgram.target_embeddings.weight.cpu().data.numpy()
save_graph_embeddings(corpus, final_embeddings, "jsonEmbeddings.json")


### Downstream classification.
graph_files = corpus.graph_fname_list
class_labels_fname = "../data/MUTAG.Labels"
graph_and_class_tuples = get_class_labels_tuples(graph_files, class_labels_fname)

embedding_fname = "jsonEmbeddings.json"
classify_scores = cross_val_accuracy(corpus_dir, corpus.extension, embedding_fname, class_labels_fname)
mean_acc, std_dev = classify_scores
print("Mean accuracy using 10 cross fold accuracy: %s with std %s" % (mean_acc, std_dev))


"""Simple script to transform TU-Dortmund Benchmark datasets into Geo2DR compatible GEXF.

This scrips helps turn datasets from the format with which TU Graph Kernel datasets are written into something we can work with.

It reads the DS_A.txt, DS_graph_indicator.txt, and DS_graph_labels.txt to create a folder of graphs in GEXF format and a graph-id to graph-classification label file.

The saved format will be 
dataset_name/dataset_name : folder containing individual gexf files of each graph.
dataset_name/dataset_name.Labels : a file denoting each gexf file to the integer class label

.. deprecated:: 0.2.1
	This script may be removed in future releases as it is succeeded by the faster
	and feature-rich dortmund_formatter.DortmundGexf class.

"""

import pickle
import sys
import os
import networkx as nx
import argparse

from collections import defaultdict
from tqdm import tqdm

parser = argparse.ArgumentParser(description='Tool to process TU graph kernel datasets')
parser.add_argument("-d", "--dataset", help="Path of the directory containing the dataset in TU format")

args = parser.parse_args()

################################################
############# Settings and Files ###############
################################################

dataset = args.dataset
graph_A_fname = "dortmund_data/" + dataset + "/" + dataset + "_A.txt"
graph_indicator_fname = "dortmund_data/" + dataset + "/" + dataset + "_graph_indicator.txt"
graph_labels_fname = "dortmund_data/" + dataset + "/" + dataset + "_graph_labels.txt"
node_labels_fname = "dortmund_data/" + dataset + "/" + dataset + "_node_labels.txt"

folder_for_graph_files = "dortmund_gexf/" + dataset

################################################
################################################
# Explanation of the TU graph kernel files

# n = total number of nodes
# m = total number of edges
# N = number of graphs

# DS_A.txt (m lines):	represents a sparse (block diagonal) adjacency matrix for all graphs, 
#						each line corresponds to (row, col) position resp. (node_id, node_id).
#						All graphs are undirected. Hence, DS_A.txt contains two entries for 
#						each edge.

# DS_graph_indicator.txt (n_lines): column vector of graph identifiers for all nodes of all graphs
#									the value in the i-th line is the graph_id of the node with 
#									node_id i

# DS_graph_labels.txt (N lines):	class labels for all graphs in the dataset, the value in the i-th
#									line is the class label of the graph with graph_id i

# DS_node_labels.txt (n lines):		column vector of node labels, the value in the i-th line is the
#									node label for node i

# There are optional files if the respective information is available:
# DS_edge_labels.txt (m lines): same size as DS_A.txt: labels for the edges in DS_A.txt
# DS_edge_attributes.txt (m lines): same size as DS_A.txt, attributes for the edges in DS_A.txt
# DS_node_attributes.txt (n lines): matrix of node attributes, the comma seperated values in the i-th line is the feature vector of the node with node_id i
# DS_graph_attributes.txt (N lines): regression values for all graphs in the datset, the value in the i-th line is the attribute of the graph with graph_id i

#################################################
##### Generating Dictionaries from files ########
#################################################
if os.path.isdir(folder_for_graph_files):
	print("#... The dataset %s already exists, closing program ...#" % (folder_for_graph_files))
else:
	print("#... Starting gexification ...#")
	# Read each line in the graph_indicator and add a node_id (of the dataset) to its corresponding 
	# graph indicated by the value
	print("#... Generating graph nodes dictionary ...#")
	graph_nodes = defaultdict(list) # each graph gets its list of nodes
	nodes = open(graph_indicator_fname).readlines()
	nodes = [int(x.strip()) for x in nodes]

	node_id = 1
	for gindex in nodes:
		graph_nodes[gindex].append(node_id)
		node_id += 1
	del(nodes)

	# Now get the edges for each graph making sure the edge nodes are actually in the graph as well
	print("#... Generating graph edges dictionary ...#") # Choke point of algorithm (but we realistically only do it once.)
	graph_edges = defaultdict(list)
	edges = open(graph_A_fname).readlines()
	edges = [x.split(",") for x in edges]
	edges = [(int(x.strip()), int(y.strip())) for x,y in edges] # nice little list of tuples

	todoedges = len(edges)
	for x,y in tqdm(edges):
		## TAKES A LOT OF TIME, TODO: parallel or smarter algo (it gets worse over time)
		for gindex in graph_nodes.keys():
			if (x in graph_nodes[gindex] and y in graph_nodes[gindex]):
				graph_edges[gindex].append((x,y))
				break # no need to continue going through graphs checking for this edge
			else:
				continue
	del(edges)

	print("#... Generating NX Graph dictionary ...#")
	# The more you know, the defaultdict is a factory pattern
	graph_nx = defaultdict()
	for gindex in tqdm(graph_edges.keys()):
		G = nx.Graph()
		for u,v in graph_edges[gindex]:
			G.add_edge(u,v)
		graph_nx[gindex] = G
	del(graph_edges)



	print("#... Relabeling nodes as necessary via attribute file ...#")
	# Our system finds unique substructures across the dataset using the node labels
	# If the nodes have labels they will be found in the node_labels_fname
	# Else if we are dealing with unlabelled graphs we will use the degree of the nodes as Labels
	if os.path.isfile(node_labels_fname):
		## This also takes time
		print("#... Relabeling nodes using %s ...#" % (node_labels_fname))
		node_att_relabel = {}
		new_node_labels = open(node_labels_fname).readlines()
		old_node_label = 1
		for new_node_label in new_node_labels:
			node_att_relabel[old_node_label] = new_node_label.strip()
			old_node_label += 1

		for gindex in graph_nx.keys():
			if gindex % 1000 == 0:
				print("Setting node label att %s" % (str(gindex)))
			nx.set_node_attributes(graph_nx[gindex], node_att_relabel, 'Label')

	else:
		print("#... Could not find node labeling file, will label by degree ...#")
		for gindex in graph_nx.keys():
			node_att_relabel = {}
			G = graph_nx[gindex]
			for node in G.nodes:
				if gindex % 50 == 0:
					print("Setting node label att %s" % (str(gindex)))
				node_att_relabel[node] = G.degree(node)
			nx.set_node_attributes(graph_nx[gindex], node_att_relabel, 'Label')

	print ("#... Physical relabeling of nodes as necessary ...#")
	for gindex in tqdm(sorted(graph_nx.keys())):
		graph_nx[gindex] = nx.convert_node_labels_to_integers(graph_nx[gindex], first_label=1)

	print ("#... Generating graph classification labels ...#")
	graph_classes = defaultdict()
	labels = open(graph_labels_fname).readlines()
	labels = [x.strip() for x in labels]

	graph_id = 1
	for label in labels:
		graph_classes[graph_id] = label
		graph_id += 1

	print("#... Generating graph dataset dictionaries completed ...#")



#################################################
### Writing to files in and making the labels ###
#################################################
	print("#... Writing to gexf files ...#")
	# We need a gexf file for each graph and a folder to put them into
	if os.path.isfile(folder_for_graph_files):
		print("Already have the folder %s" % (folder_for_graph_files))
		for gindex in sorted(graph_nx.keys()):
			graph_fname = str(gindex)+ ".gexf"
			graph_gexf_path = folder_for_graph_files + "/" + graph_fname
			nx.write_gexf(graph_nx[gindex], graph_gexf_path)

	else:
		print ("Dont have folder %s, creating and writing gexf files there" % (folder_for_graph_files))
		os.makedirs(folder_for_graph_files)
		for gindex in sorted(graph_nx.keys()):
			graph_fname = str(gindex) + ".gexf"
			graph_gexf_path = folder_for_graph_files + "/" + graph_fname
			nx.write_gexf(graph_nx[gindex], graph_gexf_path)

		# Write the labels
		print("Writing classification labels file for graphs in dataset")
		name_of_labels_file = dataset + ".Labels"
		with open(name_of_labels_file, "w") as fh:
			for gindex in sorted(graph_classes.keys()):
				graph_fname = str(gindex) + ".gexf"
				fh.write("%s %s\n" % (graph_fname, graph_classes[gindex]))
		


"""
General purpose utilities for I/O for decomposition
"""

import os
import json

def get_files(dname, extension, max_files=0):
    """
    Returns a list of strings which are all the files with 
    the given extension in a sorted manner
    """
    all_files = [os.path.join(dname, f) for f in os.listdir(dname) if f.endswith(extension)]

    for root, dirs, files, in os.walk(dname):
        for f in files:
            if f.endswith(extension):
                all_files.append(os.path.join(root,f))

    # no duplicates
    all_files = list(set(all_files))
    all_files.sort()
    if (max_files):
        return(all_files)[:max_files]
    else:
        return all_files


"""Graphlet based graph decomposition algorithm to create graph documents. 
Inspired and adapted from Yanardag and Vishwanathan "Deep Graph Kernels" [2]_.

"""

# We adopt the original license and extend it here.
# Copyright (c) 2016 Pinar Yanardag
#               2019 Paul Scherer

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


# Sytem libraries
import sys
import os
import glob
import pickle
import itertools
import random
from collections import defaultdict
from time import time

# 3rd party
import numpy as np
import networkx as nx
import pynauty # make sure to install this

# Internal
from .utils import get_files

# Random seeds from Yanardag et al.
random.seed(314124)
np.random.seed(2312312)


def load_graph(file_handle):
    """ Load the gexf format file as a nx_graph and an adjacency matrix.

    Parameters
    ----------
    file_handle : str
        path to gexf file

    Returns
    -------
    graph : networkx graph
        Networkx graph instance of input gexf file
    adj_matrix : numpy ndarray
        The adjacency matrix of the graph

    """

    graph = nx.read_gexf(file_handle)
    adj_matrix = nx.to_numpy_matrix(graph)
    return graph, adj_matrix

def get_maps(num_graphlets):
    """Load certificates created by the canonical representations of 
    graphlets into canonical maps dictionary for quick isomorphism
    check when we extract graphlets from graphs in the users dataset

    Parameters
    ----------
    num_graphlets : int
        The number of nodes in the graphlet pattern, aka the size of a graphlet pattern

    Returns
    -------
    canonical_map : dict
        A dict of Nauty certificates of the canonical representations of graphlets of the size `num_graphlets`.

    """

    data_path = os.path.join(os.path.dirname(__file__), 'canonical_maps')
    # data_path = "canonical_maps"
    with open(data_path + "/canonical_map_n%s.p"%(num_graphlets), 'rb') as handle:
        # canonical_map : {canonical string id: {"graph", "idx", "n"}}
        canonical_map = pickle.load(handle, encoding="latin1")
    return canonical_map

def get_graphlet(window, num_graphlets):
    """Compute the Nauty certificate of the graphlet in within a window of the adjacency matrix

    This function takes the upper triangle of a nxn matrix and
    computes its hash certificate using nauty. Given the parameters
    this usually involved computing the graphlet of num_graphlets 
    size

    This is used for comparison with a bank of known certificates
    as loaded in get_maps().

    Parameters
    ----------
    window : numpy ndarray
        submatrix inside the adjacency matrix of a graph
    num_graphlets: int
        the size of the graphlets to extract

    Returns
    -------
    cert : byte str
        certicate of the graphlet produced by finding its canonical representation with Nauty.

    """

    adj_mat = {idx: [i for i in list(np.where(edge)[0]) if i!=idx] for idx, edge in enumerate(window)}

    g = pynauty.Graph(number_of_vertices=num_graphlets, directed=False, adjacency_dict = adj_mat)
    cert = pynauty.certificate(g)
    return cert

def graphlet_corpus(corpus_dir, num_graphlets, samplesize):
    """Function which induces the graphlet patterns over the graphs inside a given directory.

    The main use case is for the user to input a path containing individual
    graphs of the dataset in gexf format.
    The decomposition algorithm will induce graphlet patterns for graphs
    recording the dataset/"global" vocabulary of patterns within a dictionary.
    The graph and its associated patterns (by IDs given through our hash function)
    are saved into a  <graphid>.wldr<depth> file which contains a line delimited
    list of all the substructure pattern ids.

    Parameters
    ----------
    corpus_dir : str
        path to directory containing graph files of dataset in .gexf format.
    num_graphlets : int 
        the size of the graphlet patterns to be extracted 
    samplesize : int
        the number of samples to take from a graph.

    Returns
    -------
    corpus : list of lists of str
        a list of lists, with each inner list containing all the samplesize*(1+ num_graphlets) patterns in one graph of the dataset
    vocabulary : list
        a set of the unique graphlet pattern ids
    prob_map : dict
        a map {gidx: {graphlet_idx: normalized_prob}} of normalized probabilities of a graphlet pattern appearing in a graph based on counts made in generation
    num_graphs : int
        the number of graphs in the dataset
    graph_map : dict
        a map {gidx: {graphlet_idx: count}} of the number of times a certain graphlet pattern appeared in a graph for each graph gidx in the dataset
    
    """

    fallback_map = {1: 1, 2: 2, 3: 4, 4: 8, 5: 19, 6: 53, 7: 209, 8: 1253, 9: 13599}
    canonical_map = get_maps(num_graphlets)
    vocabulary = set()
    corpus = []

    # randomly sample graphlets
    graph_map = {}
    graph_map2 = {}

    graph_files = get_files(corpus_dir, extension=".gexf")

    for gexf_fh in graph_files:
        gidx = int((os.path.basename(gexf_fh)).replace(".gexf", ""))
        nx_graph, adj_mat = load_graph(gexf_fh)
        num_nodes = len(adj_mat)

        count_map = {} # graphlet countmap
        tmp_corpus = [] # temporary corpus of graphlet patterns for a single graph
        cooccurence_corpus = [] # corpus which preserves cooccurence as in Yanardag et al.

        # Only sample graphlets if the number of nodes in the graph is larger than the
        # maximum graphlet size
        if num_nodes >= num_graphlets:
            for _ in range(samplesize):
                r =  np.random.permutation(range(num_nodes))
                for n in [num_graphlets]:
                    # "Main Graphlet"
                    # Choose a random set of num_graphlet nodes, find the graphlets of 
                    # desired size and add it to the count_map
                    window = adj_mat[np.ix_(r[0:n],r[0:n])]
                    g_type = canonical_map[str(get_graphlet(window, n), 'latin1')]
                    graphlet_idx = g_type["idx"]
                    count_map[graphlet_idx] = count_map.get(graphlet_idx, 0) + 1
                    vocabulary.add(graphlet_idx)
                    tmp_corpus.append(graphlet_idx)
                    cooccurence_corpus.append([graphlet_idx])

                    # "Co-occuring Graphlets"
                    for node in r[0:n]:
                        # select a non-overlapping window n-1 in size and one of 
                        # the nodes in the "main graphlet", to find a neighbouring graphlet
                        new_n_arr = r[n:][0:n-1]
                        r2 = np.array(list(new_n_arr) + [node])
                        window2 = adj_mat[np.ix_(r2,r2)]
                        g_type2 = canonical_map[str(get_graphlet(window2, n), 'latin1')]
                        graphlet_idx2 = g_type2["idx"]
                        count_map[graphlet_idx2] = count_map.get(graphlet_idx2, 0) + 1
                        vocabulary.add(graphlet_idx2)
                        tmp_corpus.append(graphlet_idx2)
                        cooccurence_corpus[-1].append(graphlet_idx2)

                    corpus.append(tmp_corpus)
        else:
            count_map[fallback_map[num_graphlets]] = samplesize

        # Record and save graphlet information for the one graph
        graph_map[gidx] = count_map
        save_graphlet_document(gexf_fh, gidx, num_graphlets, samplesize, cooccurence_corpus)

    # Normalise the probabilities of a graphlet in a graph.
    prob_map = {gidx: {graphlet: count/float(sum(graphlets.values())) \
        for graphlet, count in graphlets.items()} for gidx, graphlets in graph_map.items()}
    num_graphs = len(prob_map)

    return corpus, vocabulary, prob_map, num_graphs, graph_map

def save_graphlet_document(gexf_fh, gidx, num_graphlets, samplesize, cooccurence_corpus):
    """Saves the induced graphlet patterns into dataset folder, it is only used in conjunction
    with graphlet_corpus()

    Parameters
    ----------
    gexf_fh : str
        path to gexf file of graph
    gidx : int
        integer id of the graph, typically matches number in 
        gexf_fh, if using TU Dortmund benchmark datasets
    num_graphlets : int
        size of graphlet patterns induced
    samplesize : int
        number of graphlet patterns sampled from graph
    cooccurrence_corpus : list
        list of graphlet patterns induced under cooccurrence 
        rules, in this case all graphlets immediately adjacent
        to an induced graphlet pattern.

    Returns
    -------
    None : None
        The decomposition algorithm will induce graphlet patterns for graphs
        recording the dataset/"global" vocabulary of patterns within a dictionary.
        The graph and its associated patterns (by IDs given through our hash function)
        are saved into a  <graphid>.wldr<depth> file which contains a line delimited
        list of all the substructure pattern ids.


    """
    open_fname = gexf_fh + ".graphlet" + "_ng_" + str(num_graphlets) + "_ss_" + str(samplesize)
    with open(open_fname,'w') as fh:
        for graphlet_neighbourhood in cooccurence_corpus:
            sentence = str.join(" ", map(str, graphlet_neighbourhood))
            print (sentence, file=fh)


if __name__ == '__main__':
    corpus_dir = corpus_dir = "../data/dortmund_gexf/MUTAG/"
    corpus, vocabulary, prob_map, num_graphs, graph_map = graphlet_corpus(corpus_dir, num_graphlets=3, samplesize=6)
    # should result in files with 6 lines each with num_graphlets+1 items



"""PVDM model originally introduced in doc2vec paper by Le and Mikolov (2014) [6]_
Used by AWE-DD model of Anonymous Walk Embeddings by Ivanov and Burnaev (2018) [1]_


It is used with the corpus classes in cbow_data_reader which handles 
the data reading and loading. This allows construction of full PVDM
based systems. It is one of the choices of neural language model for
recreating AWE [2]_ like systems.

"""

# Author: Paul Scherer 2019

import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from torch.nn import init


class PVDM(nn.Module):
    r"""PyTorch implmentation of the PVDM as in Le and Mikolov. [6]_

    Parameters
    ----------
    num_targets : int
        The number of targets to embed. Typically the number of substructure
        patterns, but can be repurposed to be number of graphs. 
    vocab_size : int
        The size of the vocabulary; the number of unique substructure patterns
    embedding_dimension : int
        The desired dimensionality of the embeddings.

    Returns
    -------
    self : PVDM
        a torch.nn.Module of the PVDM model
    """

    def __init__(self, num_targets, vocab_size, embedding_dimension):
        super(PVDM, self).__init__()
        self.num_targets = num_targets
        self.embedding_dimension = embedding_dimension
        concat_dim = 2*embedding_dimension

        self.target_embeddings = nn.Embedding(num_targets, embedding_dimension) #D
        self.context_embeddings = nn.Embedding(vocab_size, embedding_dimension) #W
        self.output_layer = nn.Embedding(vocab_size, concat_dim) #O
        
        # Xavier initialization of weights
        initrange = 1.0 / (self.embedding_dimension)
        init.uniform_(self.target_embeddings.weight.data, -initrange, initrange)
        init.constant_(self.context_embeddings.weight.data, 0)

        self.linear1 = nn.Linear(concat_dim, 128)
        self.activation1 = nn.ReLU()
        self.linear2 = nn.Linear(128, vocab_size)
        self.activation2 = nn.LogSoftmax(dim=-1)

    def forward(self, pos_graph_emb, pos_context_target, pos_contexts, pos_negatives):
        """Forward pass in network
        
        Parameters
        ----------
        pos_graph_emb : torch.Long
            index of target graph embedding
        pos_context_target : torch.Long
            index of target subgraph pattern embedding
        pos_contexts : torch.Long
            indices of context subgraph patterns around the target subgraph embedding    
        pos_negatives : torch.Long
            indices of negatives

        Returns
        -------
        torch.float
            the negative sampling loss
        
        """

        # Be aware pos_contexts is typically several context embeddings
        # emb_target_graph = self.target_embeddings(pos_graph_emb)
        # mean_contexts_subgraphs = torch.mean(self.context_embeddings(pos_contexts), dim=1)

        # stack_target_contexts = torch.cat((emb_target_graph, mean_contexts_subgraphs), dim=1)

        # h = self.linear1(stack_target_contexts)
        # h = self.activation1(h)
        # h = self.linear2(h)
        # out = self.activation2(h)
        # return out

        # Negative sampling PV-DM
        emb_target = self.target_embeddings(pos_graph_emb)
        mean_contexts = torch.sum(self.context_embeddings(pos_contexts), dim=1)
        emb_context_target = self.output_layer(pos_context_target)
        emb_negative_targets = self.output_layer(pos_negatives)

        # Concat graph embedding with contexts embeddings
        stack_target_contexts = torch.cat((emb_target, mean_contexts), dim=1)

        objective = torch.sum(torch.mul(stack_target_contexts, emb_context_target), dim=1) # mul is elementwise multiplication
        objective = torch.clamp(objective, max=10, min=-10)
        objective = -F.logsigmoid(objective)

        neg_objective = torch.bmm(emb_negative_targets, stack_target_contexts.unsqueeze(2)).squeeze()
        neg_objective = torch.clamp(neg_objective, max=10, min=-10)
        neg_objective = -torch.sum(F.logsigmoid(-neg_objective), dim=1)

        return torch.mean(objective+neg_objective)

    def give_target_embeddings(self):
        """Return the target embeddings as a numpy matrix

        Returns
        -------
        numpy ndarray
            Numpy num_target x emb_dimension matrix of target graph embeddings
        """

        embedding = self.target_embeddings.weight.cpu().data.numpy()
        return embedding


"""This module contains algorithms useful for inducing anonymous walks in graphs.

The algorithms are adapted from the original source code by 
Ivanov and Burnaev 2018 "Anonymous Walk Embeddings" [1]_. Original reference 
implementation can be found in: https://github.com/nd7141/AWE

"""

# Author: Paul Scherer 2019
# Carry down MIT License

# Standard libraries
import sys
import os
import glob
import pickle
import itertools
import random
from collections import defaultdict
from time import time
from tqdm import tqdm

# 3rd party
import numpy as np
import networkx as nx

# Internal
from .utils import get_files

# Random seeds from reproduction
random.seed(2018)
np.random.seed(2018)


def load_graph(file_handle):
	"""Loads a nx graph object and a corresponding numpy adjacency 
	matrix given a GEXF file graph.

	Parameters
	----------
	file_handle : str
		path to gexf file

	Returns
	-------
	graph : networkx graph
		Networkx graph instance of input gexf file
	adj_matrix : numpy ndarray
		The adjacency matrix of the graph

	"""
	graph = nx.read_gexf(file_handle)
	adj_matrix = nx.to_numpy_matrix(graph)
	return graph, adj_matrix

def all_paths(paths, steps, keep_last = False):
	'''Get all possible anonymous walks of length up to steps.
	
	Parameters
	----------

	Returns
	-------
	'''
	new_paths = []
	last_step_paths = [[0, 1]]
	for i in range(2, steps+1):
		current_step_paths = []
		for j in range(i + 1):
			for walks in last_step_paths:
				if walks[-1] != j and j <= max(walks) + 1:
					new_paths.append(walks + [j])
					current_step_paths.append(walks + [j])
		last_step_paths = current_step_paths
	# filter only on n-steps walks
	if keep_last:
		new_paths = list(filter(lambda path: len(path) ==  steps + 1, new_paths))
	paths[steps] = new_paths
	return new_paths

def all_paths_edges(paths, steps, keep_last = True):
	'''Get all possible anonymous walks of length up to steps, 
	using edge labels

	'''
	new_paths = []
	last_step_paths = [[]]
	for i in range(0, steps):
		current_step_paths = []
		for j in range(i + 1):
			for walks in last_step_paths:
				if j <= max(walks + [0]) + 1:
					new_paths.append(walks + [j])
					current_step_paths.append(walks + [j])
		last_step_paths = current_step_paths
	if keep_last:
		new_paths = last_step_paths
	paths[steps] = new_paths
	return new_paths

def all_paths_nodes(paths, steps, keep_last = True):
	'''Get all possible anonymous walks of length up to steps, 
	using node labels
	
	'''
	new_paths = []
	last_step_paths = [[0]]
	for i in range(1, steps+1):
		current_step_paths = []
		for j in range(i + 1):
			for walks in last_step_paths:
				if j <= max(walks) + 1:
					new_paths.append(walks + [j])
					current_step_paths.append(walks + [j])
		last_step_paths = current_step_paths
	if keep_last:
		new_paths = last_step_paths
	paths[steps] = new_paths
	return new_paths

def all_paths_edges_nodes(paths, steps, keep_last = True):
	'''Get all possible anonymous walks of length up to steps, 
	using edge-node labels

	'''
	edge_paths = all_paths_edges(paths, steps, keep_last=keep_last)
	node_paths = all_paths_nodes(paths, steps, keep_last=keep_last)
	new_paths = []
	for p1 in edge_paths:
		for p2 in node_paths:
			if len(p2) == len(p1) + 1:
				current_path = [p2[0]]
				for ix in range(len(p1)):
					current_path.append(p1[ix])
					current_path.append(p2[ix+1])
				new_paths.append(current_path)
	paths[steps] = new_paths
	return new_paths


def create_random_walk_graph(nx_graph):
	"""Generate a random walk graph equivalent of a graph as 
	described in anonymous walk embeddings

	"""

	# get name of the label on graph edges (assume all label names are the same)
	label_name = 'weight'

	RW = nx.DiGraph()
	for node in nx_graph:
		edges = nx_graph[node]
		total = float(sum([edges[v].get(label_name, 1)
						   for v in edges if v != node]))
		for v in edges:
			if v != node:
				RW.add_edge(node, v, weight=edges[v].get(label_name, 1) / total)
	rw_graph = RW
	return rw_graph


def random_step_node(rw_graph, node):
	'''Moves one step from the current according to probabilities of 
	outgoing edges. Return next node.

	'''

	r = random.uniform(0, 1)
	low = 0
	for v in rw_graph[node]:
		p = rw_graph[node][v]['weight']
		if r <= low + p:
			return v
		low += p


def random_walk_node(rw_graph, node, steps):
	'''Creates anonymous walk from a node for arbitrary steps.
	Returns a tuple with consequent nodes.

	'''
	
	d = dict()
	d[node] = 0
	count = 1
	walk = [d[node]]
	for i in range(steps):
		v = random_step_node(rw_graph, node)
		if v not in d:
			d[v] = count
			count += 1
		walk.append(d[v])
		node = v
	return tuple(walk)


def random_walk_with_label_nodes(graph, rw_graph, node, steps):
	'''Creates anonymous walk from a node for arbitrary steps with usage of node labels.
	Returns a tuple with consequent nodes.

	'''
	
	d = dict()
	count = 0
	pattern = []
	for i in range(steps + 1):
		label = graph.nodes[node]['Label']
		if label not in d:
			d[label] = count
			count += 1
		pattern.append(d[label])
		node = random_step_node(rw_graph, node)
	return tuple(pattern)


def random_walk_with_label_edges(graph, rw_graph, node, steps):
	'''Creates anonymous walk from a node for arbitrary steps with usage of edge labels.
	Returns a tuple with consequent nodes.

	'''
	
	idx = 0
	pattern = []
	d = dict()
	for i in range(steps):
		v = random_step_node(rw_graph, node)
		label = int(graph[node][v]['Label'])
		if label not in d:
			d[label] = idx
			idx += 1
		pattern.append(d[label])
	return tuple(pattern)


def random_walk_with_label_edges_nodes(graph, rw_graph, node, steps):
	'''Creates anonymous walk from a node for arbitrary steps with usage of edge and node labels.
	Returns a tuple with consequent nodes.

	'''
	
	node_idx = 0
	edge_idx = 0
	pattern = [0]
	node_labels = dict()
	edge_labels = dict()
	for i in range(steps):
		v = random_step_node(rw_graph, node)
		node_label = graph.nodes[node]['Label']
		edge_label = int(graph[node][v]['Label'])
		if node_label not in node_labels:
			node_labels[node_label] = node_idx
			node_idx += 1
		if edge_label not in edge_labels:
			edge_labels[edge_label] = edge_idx
			edge_idx += 1
		pattern.append(node_labels[node_label])
		pattern.append(edge_labels[edge_label])
		node = v
	return tuple(pattern)


def anonymous_walk_node(graph, rw_graph, node, steps, label_setting=None):
	'''Creates anonymous walk from a node.

	'''
	
	if label_setting is None:
		return random_walk_node(rw_graph, node, steps)
	elif label_setting == 'nodes':
		return random_walk_with_label_nodes(graph, rw_graph, node, steps)
	elif label_setting == 'edges':
		return random_walk_with_label_edges(graph, rw_graph, node, steps)
	elif label_setting == 'edges_nodes':
		return random_walk_with_label_edges_nodes(graph, rw_graph, node, steps)


def anonymous_walks(graph, neighborhood_size, walk_ids, awe_length, label_setting):
	"""Generates anonymous walks for a given graph and input hyperparameters
	
	"""
	rw_graph = create_random_walk_graph(graph)
	aws = []
	for i, node in enumerate(rw_graph):
		# aw = [str(walk_ids[random_walk_node(node, steps)]) for _ in range(neighborhood_size)] # for random_walk
		aw = [str(walk_ids[anonymous_walk_node(graph, rw_graph, node, awe_length, label_setting)]) for _ in range(neighborhood_size)]  # for anonymous_walk
		aws.append(aw)
	return aws

def awe_corpus(corpus_dir, awe_length, label_setting, neighborhood_size=10, saving_graph_docs=True):
	"""Induces anonymous walks up to a supplied length across a set of graphs.

	This function extracts anonymous walks of `awe_length` across the graphs
	in a dataset into a vocabulary. The function saves a graphdoc for each graph
	which records string identifiers of the anonymous walks present in each graph

	The main use case for this function is for the user to supply a path to 
	the directory containing the .gexf file which contains .gexf files of
	each graph in a dataset. The decomposition function will produce a file 
	with the extension *.awe_<`awe_length`>_<`label_setting`>*  for each .gexf 
	file which contains the anonymous walks patterns specified by a string hash
	controlled by the vocabulary of patterns across the list of graphs being studied. 

	Parameters
	----------
	corpus_dir : str
		path to folder containing gexf graph files on which the substructure 
		patterns should be induced
	awe_length : int
		desired length of anonymous walk patterns
	label_setting : str
		information of node/edge labels that should be used; this can be None, 'nodes', 'edges', 'edges_nodes'
	neighborhood_size : int (default=10)
		the number of anonymous walks to take from a source node
	saving_graph_docs : bool (default=True)
		boolean value which dictates whether graph documents will be generated and saved. 

	Returns
	-------
	None : None
		Induces the anonymous walks of `awe_length` across the list of gexf file paths
		supplied in `corpus_dir`. Graph "documents" with the extension 
		*.awe_<`awe_length`>_<`label_setting`>* are created for each of the gexf files in the
		same directory containing string identifiers of the patterns induced in each graph.

	"""
	graph_files = get_files(corpus_dir, extension=".gexf")
	awe_corpus = {}

	# Map of possible anonymous walks
	paths = dict()
	if label_setting is None:
		all_paths(paths, awe_length, keep_last=True)
	elif label_setting == 'nodes':
		all_paths_nodes(paths, awe_length, keep_last=True)
	elif label_setting == 'edges':
		all_paths_edges(paths, awe_length, keep_last=True)
	elif label_setting == 'edges_nodes':
		all_paths_edges_nodes(paths, awe_length, keep_last=True)

	walk_ids = dict()
	for i, path in enumerate(paths[awe_length]):
		walk_ids[tuple(path)] = i

	label_suffix = ""
	if label_setting is not None:
		label_suffix = label_setting

	for gexf_fh in tqdm(graph_files):
		open_fname = gexf_fh + ".awe_" + str(awe_length) + "_" + label_suffix
		if os.path.exists(open_fname):
			continue


		gidx = int((os.path.basename(gexf_fh)).replace(".gexf", ""))
		graph, am = load_graph(gexf_fh)

		# the old write corpus rewritten as anonymous walks of a single graph
		aws_graph = anonymous_walks(graph, neighborhood_size, walk_ids, awe_length, label_setting)
		awe_corpus[gidx] = aws_graph

		if saving_graph_docs:
			# Write to the file.
			if label_suffix:
				open_fname = gexf_fh + ".awe_" + str(awe_length) + "_" + label_suffix
			else:
				open_fname = gexf_fh + ".awe_" + str(awe_length) + "_allpaths"
			with open(open_fname, 'w+') as fh:
				for aw in aws_graph:
					fh.write(' '.join(aw) + '\n')

	corpus = []
	vocabulary = set()
	graph_map = {}
	
	for gidx in sorted(awe_corpus.keys()):
		count_map = {}
		tmp_corpus = []
		for node_neighborhood_awalks in awe_corpus[gidx]:
			for pattern_str in node_neighborhood_awalks:
				vocabulary.add(pattern_str)
				count_map[pattern_str] = count_map.get(pattern_str, 0) + 1
				tmp_corpus.append(pattern_str)
		corpus.append(tmp_corpus)
		graph_map[gidx] = count_map

	# Normalise the probabilities of a graphlet in a graph.
	prob_map = {gidx: {graphlet: count/float(sum(anon_walk_patterns.values())) \
		for graphlet, count in anon_walk_patterns.items()} for gidx, anon_walk_patterns in graph_map.items()}
	num_graphs = len(graph_map)
	return corpus, vocabulary, prob_map, num_graphs, graph_map

if __name__ == '__main__':
	corpus_dir = "../data/dortmund_gexf/MUTAG/"
	awe_length = 4
	label_setting = 'nodes'
	corpus, vocabulary, prob_map, num_graphs, graph_map = awe_corpus(corpus_dir, awe_length, label_setting, saving_graph_docs=True)



"""Functions for inducing Weisfeiler Lehman graph decomposition algorithm via node relabeling
as described in Shervashidze et al. [3]_.

Based on the implementation available in the original source code of Graph2Vec [4]_
and adapted for Geo2DR https://github.com/MLDroid/graph2vec_tf which has no license

"""

# Author: Paul Scherer 2019.


import os
import glob
from time import time
import networkx as nx
from tqdm import tqdm

# Global label_to_compressed_label_map in its initial empty state 
label_to_compressed_label_map = {}

# Function to get the node label (ignore the WLk_h id before it) as an int
get_int_node_label = lambda l: int(l.split('+')[-1])

def initial_relabel(graph, node_label_attr_name="Label"): 
    """The initial relabeling of the graphs in the dataset. 

    Taking the attributed label of the node as stated by the attr 
    in the gexf file, it gives new labels to these node types
    (without regard for neighbours) hence it really is just a relabeling
    of the initial labels into our own "0+<newlabel>" format (to use 
    with WL relabeling scheme after

    Parameters
    ----------
    graph : networkx graph
        a networkx graph
    node_label_attr_name : str
        string literal of the attribute name used as a label in the 
        gexf file NB: "label" in the gexf refers to nodeID "Label"
        refers to the dataset node label
    
    Returns
    -------
    graph : networkx graph
        the same nx graph but with a new "relabel" attribute with
        the 0th wlk-h entry label

    """
    global label_to_compressed_label_map # This is the global WL hash function for compresssed labels

    nx.convert_node_labels_to_integers(graph, first_label=0)
    for node in graph.nodes(): graph.nodes[node]['relabel'] = {} # make a dictionary attribute

    # Check for previous labelings otherwise we relabel
    for node in graph.nodes():
        try:
            label = graph.nodes[node][node_label_attr_name]
        except:
            # no node label in node_label_attr_name that is specifically pulled from the gexf file so
            graph.nodes[node]['relabel'][0] = '0+0'
            continue

        if not label in label_to_compressed_label_map:
            # if no label start with 1 and increment every time a new node label is seen
            compressed_label = len(label_to_compressed_label_map)+1
            label_to_compressed_label_map[label] = compressed_label
            graph.nodes[node]['relabel'][0] = '0+' + str(compressed_label)
        else:
            # if it already has a label we just keep the same label
            graph.nodes[node]['relabel'][0] = '0+' + str(label_to_compressed_label_map[label])

    return graph

def wl_relabel(graph, it):
    """Runs an iteration of the WL relabeling algorithm, onto
    the graph using the global label_to_compressed_label_map
    
    Parameters
    ----------
    graph : networkx graph
        a networkx graph from the dataset which we want to relabel 
        (has had to be initally relabeled, ie have the graph.nodes.[node]['relabel'] attribute)
    it : int
        an int, signifiying iteration in the WL relabeling algorithm.

    Returns
    -------
    graph : networkx graph
        the input nx graph with more labels in the "relabel" attribute

    """
    global label_to_compressed_label_map # This is the global hash function for compression

    prev_iter = it - 1
    for node in graph.nodes():
        prev_iter_node_label = get_int_node_label(graph.nodes[node]['relabel'][prev_iter]) # just a int ("1") in first it 0
        node_label = [prev_iter_node_label]
        neighbours = list(nx.all_neighbors(graph, node))
        neighbourhood_label = sorted([get_int_node_label(graph.nodes[nei]['relabel'][prev_iter]) for nei in neighbours])
        node_neighbourhood_label = tuple(node_label + neighbourhood_label)
        

        if not node_neighbourhood_label in label_to_compressed_label_map:
            compressed_label = len(label_to_compressed_label_map)+1
            label_to_compressed_label_map[node_neighbourhood_label] = compressed_label
            graph.nodes[node]['relabel'][it] = str(it) + "+" + str(compressed_label)
        else:
            graph.nodes[node]['relabel'][it] = str(it) + "+" + str(label_to_compressed_label_map[node_neighbourhood_label])
    return graph


def save_wl_doc(fname,max_h,graph=None):
    """Saves the induced rooted subgraph patterns 

    Saves the subgraph sentences in format <center> <context> <context> ....
    In other words we are saving the relabelings of node from the WL algorithm 
    into a text document which can be fed into our skipgram architecture
    
    Parameters
    ----------
    fname : str
        path/filename of the graph
    max_h : int
        highest_iteration wlk_h specified (ie depth of rooted subgraph)
    graph : networkx graph
        the nx graph of the filename

    Returns
    -------
    None : None 
        The rooted subgraph patterns are saved into a text file in the 
        format <center> <context> <context> <context> ....

    """

    open_fname = fname + '.wld' + str(max_h)

    # # no need to write if it already exists
    # if os.path.isfile(open_fname):
    #     return

    # otherwise we write into the file
    with open(open_fname,'w') as fh:
        for n,d in graph.nodes(data=True):
            for it in range(0, max_h+1):
                try:
                    center = d['relabel'][it]
                except:
                    continue
                # neis_labels_prev_deg = []
                # neis_labels_next_deg = []

                # if it != 0:
                #     neis_labels_prev_deg = list(set([graph.nodes[nei]['relabel'][it-1] for nei in nx.all_neighbors(graph, n)]))
                #     neis_labels_prev_deg.sort()
                NeisLabelsSameDeg = list(set([graph.nodes[nei]['relabel'][it] for nei in nx.all_neighbors(graph,n)])) # neighbours  on iteration it basically
                # if it != max_h:
                #     neis_labels_next_deg = list(set([graph.nodes[nei]['relabel'][it+1] for nei in nx.all_neighbors(graph,n)]))
                #     neis_labels_next_deg.sort()

                # nei_list = NeisLabelsSameDeg + neis_labels_prev_deg + neis_labels_next_deg
                nei_list = NeisLabelsSameDeg
                nei_list = ' '.join(nei_list)

                sentence = center + ' ' + nei_list
                print(sentence, file=fh)
                

def wl_corpus(fnames, max_h, node_label_attr_name='Label'):
    """Induce rooted subgraph patterns using the WL node relabeling 
    algorith given list gexf files and save corresponding graph files.

    Given a set of graphs from the dataset, a maximum h for WL, and 
    the label attribute name used in the gexf files we initially relabel
    the original labels into a compliant relabeling (caesar shift for 
    ease) then perform max-h iterations of the WL relabeling algorithm 
    (1979) to create new labels which are compressed versions of the 
    rooted subgraphs for each node in the graph. These are all present
    in the nx graph objects's nodes as attributes, with the original 
    label being 'Label' and our subsequent relabelings in the "relabel"
    attribute
    
    The main use case is for the user to input a path containing individual
    graphs of the dataset in gexf format.
    The decomposition algorithm will induce substructure patterns for graphs
    recording the dataset/"global" vocabulary of patterns within a dictionary.
    The graph and its associated patterns (by IDs given through our hash function)
    are saved into a  <graphid>.wldr<depth> file which contains a line delimited
    list of all the substructure pattern ids.

    Parameters
    ----------
    fnames : list
        list of gexf file paths for the graphs in the dataset
    max_h : int
        the maximum depth of rooted subgraph pattern to induce across 
        the dataset of graphs
    node_label_attr_name : str
        string literal of the attribute name used as a label in the 
        gexf file NB: "label" in the gexf refers to nodeID "Label"
        refers to the dataset node label

    Returns
    -------
    corpus : list of lists of str
        a list of lists, with each inner list containing all the rooted subgraph patterns in one graph of the dataset
    vocabulary : list
        a set of the unique rooted subgraph pattern ids
    prob_map : dict
        a map {gidx: {wl_pattern: normalized_prob}} of normalized probabilities of a rooted subgraph pattern appearing in a graph based on counts made in generation
    num_graphs : int
        the number of graphs in the dataset
    graph_map : dict
        a map {gidx: {wl_pattern: count}} of the number of times a certain rooted subgraph pattern appeared in a graph for each graph gidx in the dataset

    None : None
        The rooted subgraph patterns are also saved into a text file for each graph in 'fnames' in the 
        format <center> <context> <context> <context> ....

    """
    global label_to_compressed_label_map
    compressed_labels_map_list = [] # list of compressed labels maps that can be used to go backwards

    # Read each graph as a networkx graph
    print('#... Loading graphs')
    graphs = [nx.read_gexf(fname) for fname in tqdm(fnames)]





    assert len(graphs) > 0, "fnames parameter does not contain valid .gexf files"
    print ('#... Loaded all the graphs')

    # Do an initial relabeling of each nxgraph g in graphs
    graphs = [initial_relabel(g, node_label_attr_name) for g in graphs]
    print ('#... initial relabeling done in')

    # Perform the Weisfeiler-Lehman Relabeling Process for h iterations (up to h depth rooted subgraphs)
    for it in range(1, max_h + 1):
        t0 = time()
        compressed_labels_map_list.append(label_to_compressed_label_map)
        label_to_compressed_label_map = {}
        graphs = [wl_relabel(g, it) for g in tqdm(graphs)]
        print ('WL iteration {} done in {} sec.'.format(it, round(time() - t0, 2)))
        print ('num of WL rooted subgraphs in iter {} is {}'.format(it, len(label_to_compressed_label_map)))

    # Save the patterns into graph documents
    for fname, g in zip(fnames, graphs):
        save_wl_doc(fname, max_h, g)

    # Match return signatures of other decomposition algorithms
    corpus = []
    vocabulary = set()
    graph_map = {}

    for fname, g in zip(fnames, graphs):
        gidx = int((os.path.basename(fname)).replace(".gexf", ""))
        tmp_corpus = []
        count_map = {}
        for n, d in g.nodes(data=True):
            for it in range(0, max_h+1):
                try:
                    pattern_at_node = d['relabel'][it]
                    vocabulary.add(pattern_at_node)
                    tmp_corpus.append(pattern_at_node)
                    count_map[pattern_at_node] = count_map.get(pattern_at_node, 0) + 1
                except:
                    continue      

                NeisLabelsSameDeg = list(set([g.nodes[nei]['relabel'][it] for nei in nx.all_neighbors(g,n)]))
                for nei_pattern in NeisLabelsSameDeg:
                    vocabulary.add(nei_pattern)
                    tmp_corpus.append(nei_pattern)
                    count_map[nei_pattern] = count_map.get(nei_pattern, 0) + 1
        
        corpus.append(tmp_corpus)
        graph_map[gidx] = count_map

    # Normalise the probabilities of a graphlet in a graph.
    prob_map = {gidx: {graphlet: count/float(sum(graphlets.values())) \
        for graphlet, count in graphlets.items()} for gidx, graphlets in graph_map.items()}
    num_graphs = len(prob_map)          

    return corpus, vocabulary, prob_map, num_graphs, graph_map

# Manual test
if __name__ == "__main__":
    ip_folder ="../data/dortmund_gexf/MUTAG"
    max_h = 2

    all_files = sorted(glob.glob(os.path.join(ip_folder, '*gexf')))
    print("Loaded %s files in total" % (str(len(all_files))))

    corpus, vocabulary, prob_map, num_graphs, graph_map = wl_corpus(all_files, max_h)


"""Functions for inducing random walk patterns within the graphs
as in Gartner et al.

Inspired by the implementation of anonymous walks by Sergey Ivanov
and Evgeny Burnaev [1]_

"""

# Author: Paul Scherer 2020

# Standard libraries
import sys
import os
import glob
import pickle
import itertools
import random
from collections import defaultdict
from time import time
from tqdm import tqdm

# 3rd party
import numpy as np
import networkx as nx

# Internal 
from .utils import get_files

# Random seeds for reproduction
random.seed(1234)
np.random.seed(1234)

def load_graph(file_handle):
	"""Loads a nx graph object and a corresponding numpy adjacency 
	matrix given a GEXF file graph.

	Parameters
	----------
	file_handle : str
		path to gexf file

	Returns
	-------
	graph : networkx graph
		Networkx graph instance of input gexf file
	adj_matrix : numpy ndarray
		The adjacency matrix of the graph

	"""
	graph = nx.read_gexf(file_handle)
	adj_matrix = nx.to_numpy_matrix(graph)
	return graph, adj_matrix

def random_step_node(rw_graph, node):
	'''Moves one step from the current according to probabilities of 
	outgoing edges. Return next node.

	'''

	r = random.uniform(0, 1)
	low = 0
	for v in rw_graph[node]:
		p = rw_graph[node][v]['weight']
		if r <= low + p:
			return v
		low += p

def random_walk_with_label_nodes(graph, rw_graph, node, steps):
	'''Creates anonymous walk from a node for arbitrary steps with usage of node labels.
	Returns a tuple with consequent nodes.

	'''
	
	d = dict()
	count = 0
	pattern = []
	for i in range(steps + 1):
		label = graph.nodes[node]['Label']
		if label not in d:
			d[label] = count
			count += 1
		pattern.append(d[label])
		node = random_step_node(rw_graph, node)
	return tuple(pattern)

def create_random_walk_graph(nx_graph):
	"""Generate a random walk graph equivalent of a graph as 
	described in anonymous walk embeddings
	"""

	# get name of the label on graph edges (assume all label names are the same)
	label_name = 'weight'

	RW = nx.DiGraph()
	for node in nx_graph:
		edges = nx_graph[node]
		total = float(sum([edges[v].get(label_name, 1)
						   for v in edges if v != node]))
		for v in edges:
			if v != node:
				RW.add_edge(node, v, weight=edges[v].get(label_name, 1) / total)
	rw_graph = RW
	return rw_graph


def random_walks_graph(nx_graph, walk_length, neighborhood_size=0):
	"""Perform a random walk with cooccurring "neighbouring" walks
	starting from the same node for all the nodes in the graph.
	"""
	neigh_size = neighborhood_size + 1 # there is at least one random walk per node
	rw_graph = create_random_walk_graph(nx_graph)
	walks = []
	for i, node in enumerate(rw_graph):
		rw = [str(random_walk_with_label_nodes(nx_graph, rw_graph, node, walk_length)) for _ in range(neighborhood_size)]  # for anonymous_walk
		walks.append(rw)
	return walks

def save_rw_doc(gexf_fh, coocurrence_corpus, walk_length):
	"""Saves a shortest path graph doc with the extension .spp
	
	Parameters
	----------
	gexf_fh : str
		sdf
	cooccurrence_corpus : list
		list of lists containing cooccuring shortest paths 
		(ie those starting from the same starting node)
	walk_length : int
		length of the walk

	Returns
	-------
	None : None
		Saves the induced shortest path patterns into a graph document for the graph specified
		in the `gexf_fh` 

	"""

	open_fname = gexf_fh + ".rw"  + str(walk_length)
	# if os.path.isfile(open_fname):
	# 	return
	with open(open_fname,'w') as fh:
		for spp_neighbourhood in coocurrence_corpus:
			sentence = str.join(" ", map(str, spp_neighbourhood))
			print (sentence, file=fh)

def rw_corpus(corpus_dir, walk_length, neighborhood_size, saving_graph_docs=True):
	"""Induces random walks up to a desired length across the input set of graphs
	
	Parameters
	----------
	corpus_dir : str
		path to directory containing graph files
	walk_length : int
		desired length of the random walk
	neighborhood_size : int
		number of cooccuring walks to find per walk source (node)
	saving_graph_docs : bool
		boolean on whether the graph documents should be generated or not

	Returns
	-------
	None : None
		Induces the random walks of `walk_length` across the list of graphs files
		supplied in `corpus_dir` 
	"""

	graph_files = get_files(corpus_dir, extension=".gexf")
	vocabulary = set()
	count_map = {}
	graph_map = {}
	corpus = []
	pattern_to_short_id = {}
	short_id_to_patern = {}
	pattern_id = 0

	for gexf_fh in tqdm(graph_files):
		gidx = int((os.path.basename(gexf_fh)).replace(".gexf", ""))
		graph, am = load_graph(gexf_fh)

		tmp_corpus = []
		count_map[gidx] = {}
		cooccurence_corpus = []
		
		# Induce random walks across the single graph
		rws_graph = random_walks_graph(graph, walk_length, neighborhood_size)
		for node_walks in rws_graph:
			paths_cooccurence = []
			for source_wlk in node_walks:
				# make a shorter label for each pattern
				if source_wlk in pattern_to_short_id:
					label = pattern_to_short_id[source_wlk]
				else:
					pattern_to_short_id[source_wlk] = str(pattern_id)
					short_id_to_patern[str(pattern_id)] = source_wlk
					label = pattern_to_short_id[source_wlk]
					pattern_id += 1

				tmp_corpus.append(label)
				paths_cooccurence.append(label)
				count_map[gidx][label] = count_map[gidx].get(label,0)+1
				vocabulary.add(label)
			cooccurence_corpus.append(paths_cooccurence)

		corpus.append(tmp_corpus)
		if saving_graph_docs:
			save_rw_doc(gexf_fh, cooccurence_corpus, walk_length)

	graph_map = count_map
	prob_map = {gidx: {path: count/float(sum(paths.values())) \
		for path, count in paths.items()} for gidx, paths in count_map.items()}

	num_graphs = len(count_map)
	return corpus, vocabulary, prob_map, num_graphs, graph_map

if __name__ == "__main__":
	corpus_dir = "../data/dortmund_gexf/MUTAG/"
	corpus, vocabulary, prob_map, num_graphs, graph_map = rw_corpus(corpus_dir, 4, 5)


"""Module containing various functions for classification (on top of the learned embeddings)
mainly useful for providing convenience functions on common benchmark classification methods

"""
import json

# Sklearn SVC (for "fair" comparison with existing methods)
from sklearn.feature_extraction.text import CountVectorizer
from sklearn.preprocessing import Normalizer
from sklearn.model_selection import train_test_split
from sklearn.metrics import accuracy_score, classification_report, precision_recall_fscore_support
from sklearn.svm import SVC, LinearSVC
from sklearn.model_selection import GridSearchCV

from random import randint
import numpy as np
import logging

from .utils import get_files, get_class_labels

def linear_svm_classify(X_train, X_test, Y_train, Y_test):
	"""Utility function for quickly performing Scikit Learn GridSearchCV over a linear SVM 
	with 10 fold CrossVal given the train test splits
	
	Parameters
	----------
	X_train : numpy ndarray
		training feature vectors
	X_test : numpy ndarray
		testing feature vectors
	Y_train : numpy ndarray
		training set labels
	Y_test : numpy ndarray
		test set labels

	Returns
	-------
	tuple
		tuple with accuracy, precision, recall, fbeta_score as applicable
	"""
	params = {'C':[0.001, 0.01, 0.1, 1, 10, 100, 1000]}
	if len(set(Y_train)) == 2:
		classifier = GridSearchCV(LinearSVC(max_iter=100000000), params, cv=10, scoring='f1', verbose=1, n_jobs=-1)
	else:
		classifier = GridSearchCV(LinearSVC(max_iter=100000000), params, cv=10, scoring='f1_weighted', verbose=1, n_jobs=-1)
	classifier.fit(X_train, Y_train)
	logging.info('best classifier models hyperparameters', classifier.best_params_)

	Y_pred = classifier.predict(X_test)

	acc = accuracy_score(Y_test, Y_pred)
	logging.info('Linear SVM accuracy: {}'.format(acc))

	report = classification_report(Y_test, Y_pred)
	logging.info(report)

	precision, recall, fbeta_score, support = precision_recall_fscore_support(Y_test, Y_pred)

	return (acc, precision, recall, fbeta_score)

def rbf_svm_classify(X_train, X_test, Y_train, Y_test):
	"""Utility function for quickly performing Scikit Learn 
	GridSearchCV over a rbf kernel SVM with 10 fold CrossVal 
	given the train test splits
	
	Parameters
	----------
	X_train : numpy ndarray
		training feature vectors
	X_test : numpy ndarray
		testing feature vectors
	Y_train : numpy ndarray
		training set labels
	Y_test : numpy ndarray
		test set labels

	Returns
	-------
	tuple
		tuple with accuracy, precision, recall, fbeta_score as applicable
	"""

	params = {'C':[0.001, 0.01, 0.1, 1, 10, 100, 1000]}
	if len(set(Y_train)) == 2:
		classifier = GridSearchCV(SVC(gamma="scale"), params, cv=10, scoring='f1', verbose=1, n_jobs=-1)
	else:
		classifier = GridSearchCV(SVC(gamma="scale"), params, cv=10, scoring='f1_weighted', verbose=1, n_jobs=-1)
	classifier.fit(X_train, Y_train)

	Y_pred = classifier.predict(X_test)
	acc = accuracy_score(Y_test, Y_pred)
	precision, recall, fbeta_score, support = precision_recall_fscore_support(Y_test, Y_pred)

	return (acc, precision, recall, fbeta_score)


def perform_classification(corpus_dir, extension, embedding_fname, class_labels_fname):
	"""Perform classification over the graph files of dataset given they have corresponding
	embeddings in the saved embedding file and class labels

	Parameters
	----------
	corpus_dir : str 
		folder containing graphdoc files
	extension : str 
		extension of the graphdoc files
	embedding_fname : str 
		file containing embeddings
	class_labels_fname : str 
		files containing labels of each graph
	
	Returns
	-------
	tuple
		tuple with accuracy, precision, recall, fbeta_score as applicable

	"""

	wlk_files = get_files(corpus_dir, extension)

	Y = np.array(get_class_labels(wlk_files, class_labels_fname))
	logging.info('Y (label) matrix shape: {}'.format(Y.shape))

	seed = randint(0,1000)
	with open(embedding_fname, 'r') as fh:
		graph_embedding_dict = json.load(fh)
	X = np.array([graph_embedding_dict[fname] for fname in wlk_files])

	X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size = 0.1, random_state = seed)
	logging.info('Training and Test Matrix Shapes: {}. {}. {}. {} '.format(X_train.shape , X_test.shape, Y_train.shape, Y_test.shape))

	scores = rbf_svm_classify(X_train, X_test, Y_train, Y_test)
	return scores

def cross_val_accuracy(corpus_dir, extension, embedding_fname, class_labels_fname, cv=10, mode=None):
	"""
	Performs 10 (default) fold cross validation, returns the mean accuracy and associated 
	standard deviation
	
	Parameters
	----------
	corpus_dir : str
		folder containing graphdoc files
	extension : str
		extension of the graphdoc files
	embedding_fname : str
		file containing embeddings
	class_labels_fname : str
		files containing labels of each graph
	cv : int
		integer stating number of folds and therefore experiments to carry out

	Returns
	-------
	tuple : (acc, std)
		tuple containing the mean accuracies of performing 10 fold cross validation 10 times.
		This gives a better picture of usual performance expected performance in a Monte 
		Carlo fashion instead of presenting just best performance.

	"""
	# our accuracies
	acc_results = []
	wlk_files = get_files(corpus_dir, extension)
	Y = np.array(get_class_labels(wlk_files, class_labels_fname))

	for i in range(cv):
		seed = randint(0,1000)
		with open(embedding_fname, 'r') as fh:
			graph_embedding_dict = json.load(fh)
		X = np.array([graph_embedding_dict[fname] for fname in wlk_files])

		X_train, X_test, Y_train, Y_test = train_test_split(X, Y, test_size = 0.1, random_state = seed)

		if mode == "linear":
			scores = linear_svm_classify(X_train, X_test, Y_train, Y_test)
		else:
			scores = rbf_svm_classify(X_train, X_test, Y_train, Y_test)
		
		acc_results.append(scores[0])

	return np.mean(acc_results), np.std(acc_results)

def cross_val_accuracy_rbf_bag_of_words(P, y_ids, cv=10):
	r"""cv times Monte Carlo experimentation of 10 fold cross validation, used on
	given dataset matrix returns overall mean accuracy and associated standard deviation. 
	Terminology and method name will be updated in future version to address overloading 
	term and generalizability of function.
	
	Parameters
	----------
	P : numpy ndarray
		a obs x num_features matrix showing dataset
	y_ids : numpy ndarray
		numpy 1 x obs array of class labels for the rows of `P`
	cv : int (default=10)
		overloaded term of monte carlo restarts of the SVM evaluation over 10 fold CV
	
	Returns
	-------
	tuple : (acc, std)
		tuple containing the mean accuracies of performing 10 fold cross validation 10 times.
		This gives a better picture of usual performance expected performance in a Monte 
		Carlo fashion instead of presenting just best performance.

	"""
	acc_results = []
	Y = np.array(y_ids)
	seeds = range(cv)
	for i in range(cv):
		seed = seeds[i]
		X_train, X_test, Y_train, Y_test = train_test_split(P, Y, test_size = 0.1, random_state = seed)
		scores = rbf_svm_classify(X_train, X_test, Y_train, Y_test)
		acc_results.append(scores[0])
	return np.mean(acc_results), np.std(acc_results)


"""Gexifier for TU Dortmund graph kernel based datasets.

"""

import pickle
import sys
import os
import networkx as nx
from collections import defaultdict
from tqdm import tqdm


################################################
################################################
# Explanation of the TU graph kernel files

# n = total number of nodes
# m = total number of edges
# N = number of graphs

# DS_A.txt (m lines):	represents a sparse (block diagonal) adjacency matrix for all graphs, 
#						each line corresponds to (row, col) position resp. (node_id, node_id).
#						All graphs are undirected. Hence, DS_A.txt contains two entries for 
#						each edge.

# DS_graph_indicator.txt (n_lines): column vector of graph identifiers for all nodes of all graphs
#									the value in the i-th line is the graph_id of the node with 
#									node_id i

# DS_graph_labels.txt (N lines):	class labels for all graphs in the dataset, the value in the i-th
#									line is the class label of the graph with graph_id i

# DS_node_labels.txt (n lines):		column vector of node labels, the value in the i-th line is the
#									node label for node i

# There are optional files if the respective information is available:
# DS_edge_labels.txt (m lines): same size as DS_A.txt: labels for the edges in DS_A.txt
# DS_edge_attributes.txt (m lines): same size as DS_A.txt, attributes for the edges in DS_A.txt
# DS_node_attributes.txt (n lines): matrix of node attributes, the comma seperated values in the i-th line is the feature vector of the node with node_id i
# DS_graph_attributes.txt (N lines): regression values for all graphs in the datset, the value in the i-th line is the attribute of the graph with graph_id i


class DortmundGexf(object):
	"""A class which reads TU Dortmund style datasets and processes them
	into a corresponding set of .gexf graphs and an associated .Labels file
	
	This class helps turn datasets from the format with which 
	TU Graph Kernel datasets are written into Gexf datasets,
	which Geo2DR can work with.

	It reads the DS_A.txt, DS_graph_indicator.txt, and DS_graph_labels.txt 
	to create a folder of graphs in GEXF format and a graph-id to 
	graph-classification label file.

	The saved format will be 
	dataset_name/dataset_name/<name>.gexf : folder containing individual gexf files of each graph.
	dataset_name/dataset_name.Labels : a file denoting each gexf file to the integer class label

	See tu_gexifier for a more basic script based version. This class version
	will also contain various metadata about the dataset which may be useful
	for downstream decomposition algorithms and other analysis
	
	Parameters
	----------
	dataset : str
		string name of directory containing dataset, eg. "MUTAG".
	path_to_dataset : str
		path to directory containing directory of dataset.
	output_dir_for_graph_files : str
		path to where new dataset and labels file will be saved.
	"""
	def __init__(self, dataset, path_to_dataset, output_dir_for_graph_files):
		super(DortmundGexf, self).__init__()
		self.dataset = dataset
		# TODO Do some regex dark magicks 
		self.graph_A_fname = path_to_dataset + dataset + "/" + dataset + "_A.txt"
		self.graph_indicator_fname = path_to_dataset + dataset + "/" + dataset + "_graph_indicator.txt"
		self.graph_labels_fname = path_to_dataset + dataset + "/" + dataset + "_graph_labels.txt"
		self.node_labels_fname = path_to_dataset + dataset + "/" + dataset + "_node_labels.txt"
		self.output_dir_for_graph_files = output_dir_for_graph_files
		self.folder_for_graph_files = output_dir_for_graph_files + dataset

	def format_dataset(self):
		"""Method which formats supplied TU-Dortmund formatted dataset into 
		GEXF format compatible with other geometric2dr modules

		Returns
		-------
		None
			The formatted dataset will be saved in `output_dir_for_graph_files` 
			with the format described above
		"""  
		if os.path.isdir(self.folder_for_graph_files):
			print("#... The dataset %s already exists, closing program ...#" % (self.folder_for_graph_files))
		else:
			print("#... Starting gexification ...#")
			# Read each line in the graph_indicator and add a node_id (of the dataset) to its corresponding 
			# graph indicated by the value
			print("#... Generating graph nodes dictionary ...#")
			graph_nodes = defaultdict(list) # each graph gets its list of nodes
			nodes = open(self.graph_indicator_fname).readlines()
			nodes = [int(x.strip()) for x in nodes]

			nodes_to_graph = {}

			node_id = 1
			for gindex in nodes:
				graph_nodes[gindex].append(node_id)
				nodes_to_graph[node_id] = gindex
				node_id += 1
			del(nodes)

			# Now get the edges for each graph making sure the edge nodes are actually in the graph as well
			print("#... Generating graph edges dictionary ...#") # Choke point of algorithm (but we realistically only do it once.)
			edges = open(self.graph_A_fname).readlines()
			edges = [x.split(",") for x in edges]
			edges = [(int(x.strip()), int(y.strip())) for x,y in edges] # nice little list of tuples

			todoedges = len(edges)

			# new version
			graph_edges = defaultdict(list)
			for x,y in tqdm(edges):
				x_gindex = nodes_to_graph[x]
				y_gindex = nodes_to_graph[y]
				if x_gindex == y_gindex:
					graph_edges[x_gindex].append((x,y))
			del(edges)

			print("#... Generating NX Graph dictionary ...#")
			# The more you know, the defaultdict is a factory pattern
			graph_nx = defaultdict()
			for gindex in tqdm(graph_edges.keys()):
				G = nx.Graph()
				for u,v in graph_edges[gindex]:
					G.add_edge(u,v)
				graph_nx[gindex] = G
			del(graph_edges)

			print("#... Relabeling nodes as necessary via attribute file ...#")
			# Our system finds unique substructures across the dataset using the node labels
			# If the nodes have labels they will be found in the node_labels_fname
			# Else if we are dealing with unlabelled graphs we will use the degree of the nodes as Labels
			if os.path.isfile(self.node_labels_fname):
				## This also takes time
				print("#... Relabeling nodes using %s ...#" % (self.node_labels_fname))
				node_att_relabel = {}
				new_node_labels = open(self.node_labels_fname).readlines()
				old_node_label = 1
				for new_node_label in new_node_labels:
					node_att_relabel[old_node_label] = new_node_label.strip()
					old_node_label += 1

				for gindex in graph_nx.keys():
					if gindex % 1000 == 0:
						print("Setting node label att %s" % (str(gindex)))
					nx.set_node_attributes(graph_nx[gindex], node_att_relabel, 'Label')

			else:
				print("#... Could not find node labeling file, will label by degree ...#")
				for gindex in graph_nx.keys():
					node_att_relabel = {}
					G = graph_nx[gindex]
					for node in G.nodes:
						if gindex % 50 == 0:
							print("Setting node label att %s" % (str(gindex)))
						node_att_relabel[node] = G.degree(node)
					nx.set_node_attributes(graph_nx[gindex], node_att_relabel, 'Label')

			print ("#... Physical relabeling of nodes as necessary ...#")
			for gindex in tqdm(sorted(graph_nx.keys())):
				graph_nx[gindex] = nx.convert_node_labels_to_integers(graph_nx[gindex], first_label=1)

			print ("#... Generating graph classification labels ...#")
			graph_classes = defaultdict()
			labels = open(self.graph_labels_fname).readlines()
			labels = [x.strip() for x in labels]

			graph_id = 1
			for label in labels:
				graph_classes[graph_id] = label
				graph_id += 1

			print("#... Generating graph dataset dictionaries completed ...#")

			#################################################
			### Writing to files in and making the labels ###
			#################################################
			print("#... Writing to gexf files ...#")
			# We need a gexf file for each graph and a folder to put them into
			if os.path.isfile(self.folder_for_graph_files):
				print("Already have the folder %s" % (self.folder_for_graph_files))
				for gindex in sorted(graph_nx.keys()):
					graph_fname = str(gindex)+ ".gexf"
					graph_gexf_path = self.folder_for_graph_files + "/" + graph_fname
					nx.write_gexf(graph_nx[gindex], graph_gexf_path)

			else:
				print ("Dont have folder %s, creating and writing gexf files there" % (self.folder_for_graph_files))
				os.makedirs(self.folder_for_graph_files)
				for gindex in sorted(graph_nx.keys()):
					graph_fname = str(gindex) + ".gexf"
					graph_gexf_path = self.folder_for_graph_files + "/" + graph_fname
					nx.write_gexf(graph_nx[gindex], graph_gexf_path)

				# Write the labels
				print("Writing classification labels file for graphs in dataset")
				name_of_labels_file = self.output_dir_for_graph_files + self.dataset + ".Labels"
				with open(name_of_labels_file, "w") as fh:
					for gindex in sorted(graph_classes.keys()):
						graph_fname = str(gindex) + ".gexf"
						fh.write("%s %s\n" % (graph_fname, graph_classes[gindex]))
					

if __name__ == '__main__':
	gexifier = DortmundGexf("MUTAG", "dortmund_data/", "/tmp/")
	gexifier.format_dataset()


"""CBOW model with negative sampling as in Mikolov et al. [5]_. 

It is used with the corpus classes in cbow_data_reader which handles 
the data reading and loading. This allows construction of full CBOW
based systems. It is one of the choices of neural language model for
recreating DGK [2]_ like systems.

"""

# Author: Paul Scherer

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

class Cbow(nn.Module):
	"""Pytorch implementation of the CBOW architecture with negative sampling
	as in Mikolov et al. [5]_

	This is used in DGK models for example to learn embeddings
	of substructures for downstream graph kernel definitions.

	Parameters
	----------
	num_targets : int
		The number of targets to embed. Typically the number of substructure
		patterns, but can be repurposed to be number of graphs. 
	vocab_size : int
		The size of the vocabulary; the number of unique substructure patterns
	embedding_dimension : int
		The desired dimensionality of the embeddings.

	Returns
	-------
	self : CBow 
		a torch.nn.Module of the CBOW model
	"""
	def __init__(self, num_targets, vocab_size, embedding_dimension):
		super(Cbow, self).__init__()
		self.num_targets = num_targets
		self.embedding_dimension = embedding_dimension
		self.vocab_size = vocab_size
		
		self.target_embeddings = nn.Embedding(num_targets, embedding_dimension) #D
		self.context_embeddings = nn.Embedding(vocab_size, embedding_dimension) #W
		self.output_layer = nn.Embedding(vocab_size, embedding_dimension) #O

		# Xavier initialization of weights
		initrange = 1.0 / (self.embedding_dimension)
		init.uniform_(self.target_embeddings.weight.data, -initrange, initrange)
		init.constant_(self.context_embeddings.weight.data, 0)

		self.linear1 = nn.Linear(embedding_dimension, 128)
		self.activation1 = nn.ReLU()
		self.linear2 = nn.Linear(128, vocab_size)
		self.activation2 = nn.LogSoftmax(dim=-1)

	def forward(self, pos_target, pos_contexts, pos_negatives):
		"""Forward pass in network
		
		Parameters
		----------
		pos_target : torch.Long
			index of target embedding
		pos_contexts : torch.Long
			indices of context embeddings
		pos_negatives : torch.Long
			indices of negatives

		Returns
		-------
		torch.float
			the negative sampling loss
		
		"""
		# Be aware pos_contexts is typically several context embeddings
		# emb_target_graph = self.context_embeddings(pos_graph_emb)
		# mean_contexts_subgraphs = torch.mean(self.context_embeddings(pos_contexts), dim=1)

		# stack_target_contexts = torch.cat((emb_target_graph, mean_contexts_subgraphs), dim=1)

		# h = self.linear1(stack_target_contexts)
		# h = self.activation1(h)
		# h = self.linear2(h)
		# out = self.activation2(h)
		# return out

		# Negative sampling CBOW
		mean_contexts = torch.sum(self.context_embeddings(pos_contexts), dim=1)
		emb_context_target = self.output_layer(pos_target)
		emb_negative_targets = self.output_layer(pos_negatives)


		objective = torch.sum(torch.mul(mean_contexts, emb_context_target), dim=1) # mul is elementwise multiplication
		objective = torch.clamp(objective, max=10, min=-10)
		objective = -F.logsigmoid(objective)

		neg_objective = torch.bmm(emb_negative_targets, mean_contexts.unsqueeze(2)).squeeze()
		neg_objective = torch.clamp(neg_objective, max=10, min=-10)
		neg_objective = -torch.sum(F.logsigmoid(-neg_objective), dim=1)

		return torch.mean(objective+neg_objective)

	def give_target_embeddings(self):
		"""Return the target embeddings as a numpy matrix

		Returns
		-------
		numpy ndarray
		    Numpy vocab_size x emb_dimension matrix of substructure pattern embeddings
		"""

		embedding = self.target_embeddings.weight.cpu().data.numpy()
		return embedding


"""General Skipgram model with negative sampling originally introduced by word2vec paper
Mikolov et al [5]_. Used by DGK [2]_ and Graph2Vec [4]_ to learn substructure and graph level embeddings 

It is used by the SkipgamCorpus and PVDBOWCorpus to build complete Skipgram and PVDBOW systems respectively.
SkipgramCorpus and PVDBOWCorpus are found in skipgram_data_reader and pvdbow_data_reader modules respectively

Author: Paul Scherer
"""

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn import init

class Skipgram(nn.Module):
	"""Pytorch implementation of the skipgram with negative 
	sampling as in Mikolov et al. [5]_

	Based on the inputs it can be used as the skipgram described in the original Word2Vec paper [5]_ , or as 
	Doc2Vec (PV-DBOW) in Le and Mikolov [6]_

	Parameters
	----------
	num_targets : int
		The number of targets to embed. Typically the number of substructure
		patterns, but can be repurposed to be number of graphs. 
	vocab_size : int
		The size of the vocabulary; the number of unique substructure patterns
	embedding_dimension : int
		The desired dimensionality of the embeddings.

	Returns
	-------
	self : Skipgram 
		a torch.nn.Module of the Skipgram model
	"""

	def __init__(self, num_targets, vocab_size, embedding_dimension):
		super(Skipgram, self).__init__()
		self.num_targets = num_targets
		self.embedding_dimension = embedding_dimension
		self.target_embeddings = nn.Embedding(num_targets, embedding_dimension, sparse=True) # its a tensor with a lookup function overloading the embedding(word) oart
		self.context_embeddings = nn.Embedding(vocab_size, embedding_dimension, sparse=True)

		# Xavier initialization of weights
		initrange=1.0/(self.embedding_dimension) 
		init.uniform_(self.target_embeddings.weight.data, -initrange, initrange)
		init.constant_(self.context_embeddings.weight.data,0)

	def forward(self, pos_target, pos_context, neg_context):
		"""Forward pass in network
		
		Parameters
		----------
		pos_target : torch.Long
			index of target embedding
		pos_context : torch.Long
			index of context embedding
		neg_context : torch.Long
			index of negative
		
		Returns
		-------
		torch.float
			the negative sampling loss
			
		"""
		emb_target = self.target_embeddings(pos_target)
		emb_context = self.context_embeddings(pos_context)
		emb_neg_context = self.context_embeddings(neg_context)

		objective = torch.sum(torch.mul(emb_target, emb_context), dim=1) # mul is elementwise multiplication
		objective = torch.clamp(objective, max=10, min=-10)
		objective = -F.logsigmoid(objective)

		neg_objective = torch.bmm(emb_neg_context, emb_target.unsqueeze(2)).squeeze()
		neg_objective = torch.clamp(neg_objective, max=10, min=-10)
		neg_objective = -torch.sum(F.logsigmoid(-neg_objective), dim=1)

		return torch.mean(objective+neg_objective)

	def give_target_embeddings(self):
		embedding = self.target_embeddings.weight.cpu().data.numpy()
		return embedding


from setuptools import setup, find_packages

with open('README.md') as f:
	long_description = f.read()

setup(name="geometric2dr",
      version='0.5.0',
      description="Geo2DR: Library to build distributed representations of graphs",
      long_description = long_description,
      long_description_content_type="text/markdown",
      url="https://github.com/paulmorio/geo2dr",
      author="Paul Scherer",
      author_email="paul.geo2dr@gmail.com",
      license="MIT",
      # packages=["geometric2dr"],
      install_requires=['numpy', 'scikit-learn', 'networkx', 'tqdm', 'ipython', 'torch', 'torchvision'],
      test_suite='nose.collector',
      tests_require=['nose', 'nose-cover3'],
      include_package_data=True,
      packages=find_packages(),
      classifiers=['Topic :: Scientific/Engineering :: Artificial Intelligence',
      				'License :: OSI Approved :: MIT License',
      				'Natural Language :: English',
      				'Programming Language :: Python :: 3.6',
      				'Operating System :: OS Independent'],
      zip_safe=False
      )



"""General purpose utilities for I/O

Currently Includes:
Functions for getting all files in a directory with a given extension
Saving graph embeddings into a JSON format
Generating a dictionary matching graph files with classification labels

"""

import os
import json

def get_files(dname, extension, max_files=0):
    """Returns a list of strings which are all the files with 
    the given extension in a sorted manner

    Parameters
    ----------
    dname : str
        directory with files
    extension : str
        string denoting which extension should be matched in search for files
    max_files : int (default=0)
        the maximum number of files to get, the default of 0 means all files

    Returns
    -------
    all_files : list
        list of all files matching extension inside the directory dname

    """
    all_files = [os.path.join(dname, f) for f in os.listdir(dname) if f.endswith(extension)]

    for root, dirs, files, in os.walk(dname):
        for f in files:
            if f.endswith(extension):
                all_files.append(os.path.join(root,f))

    # no duplicates
    all_files = list(set(all_files))
    all_files.sort()
    if (max_files):
        return(all_files)[:max_files]
    else:
        return all_files

def save_graph_embeddings(corpus, final_embeddings, opfname):
    """Saves the trained embeddings of a corpus into a dictionary
    and saves this into a json file on the path given by opfname

    Parameters
    ----------
    corpus : corpus
        any corpus class such as `PVDBOWCorpus`
    final_embeddings : numpy ndarray
        matrix of target embeddings to be saved
    opfname : str
        path to file where embeddings should be saved in json format (extension optional in Unix)

    Returns
    -------
    None
        embeddings will be saved into path denoted by `opfname`
    """

    dict_to_save = {}
    for i in range(len(final_embeddings)):
        graph_fname = corpus._id_to_graph_name_map[i]
        graph_embedding = final_embeddings[i,:].tolist()
        dict_to_save[graph_fname] =  graph_embedding

    with open(opfname, 'w') as filehandler:
        json.dump(dict_to_save, filehandler, indent=4)

def save_subgraph_embeddings(corpus, final_embeddings, opfname):
    """Save the embeddings along with a map to the patterns and the corpus
    
    Parameters
    ----------
    corpus : corpus
        a corpus class such as SkipgramCorpus
    final_embeddings : numpy ndarray
        matrix of target embeddings to be saved
    opfname : str
        path to file where embeddings should be saved in json format

    Returns
    -------
    None
        embeddings will be saved into path denoted by `opfname`

    """
    dict_to_save = {}
    for i in range(len(final_embeddings)):
        subgraph_name = corpus._id_to_subgraph_map[i]
        subgraph_embedding = final_embeddings[i,:].tolist()
        dict_to_save[subgraph_name] =  subgraph_embedding

    with open(opfname, 'w') as filehandler:
        json.dump(dict_to_save, filehandler, indent=4)

def get_class_labels(graph_files, class_labels_fname):
    """Given the list of graph files (as in get_files) and 
    path of the associated class labels returns the list 
    of labels associated with each graph file in graph_files

    Parameters
    ----------
    graph_files : list
        list of paths to graph_files
    class_labels_fname : str
        path to class labels file (.Labels typically) with file names in `graph_files`

    Returns
    -------
    labels : list
        list of class labels for corresponding to graph files in `graph_files`

    """
    graph_to_class_label_map = {l.split()[0].split('.')[0]: int(l.split()[1].strip()) for l in open (class_labels_fname)}
    labels = [graph_to_class_label_map[os.path.basename(g).split('.')[0]] for g in graph_files]
    return labels

def get_class_labels_tuples(graph_files, class_labels_fname):
    """Returns list of tuples associating each of the graph files 
    to their classification labels

    Parameters
    ----------
    graph_files : list
        list of paths to graph_files
    class_labels_fname : str
        path to class labels file (.Labels typically) with file names in `graph_files`

    Returns
    -------
    labels : list
        list of tuples (base_name_of_graph_file, class_label)

    """
    graph_to_class_label_map = {l.split()[0].split('.')[0]: int(l.split()[1].strip()) for l in open (class_labels_fname)}
    labels = []
    for g in graph_files:
        g_num = os.path.basename(g).split('.')[0]
        labels.append((int(g_num), graph_to_class_label_map[os.path.basename(g).split('.')[0]]))
    return labels


def get_kernel_matrix_row_idx_with_class(corpus, extension, graph_files, class_labels_fname):
    """Returns two arrays, the first is an list of integers each referencing a row in
    a kernel matrix and thereby a kernel vector corresponding to one of the graphs 
    in the dataset, the second is a list of class labels whose value is the classification
    of the graph in the same index of the first

    Parameters
    ----------
    corpus : corpus
        a corpus instance (such as SkipgramCorpus)
    extension : str
        extension of graph document under study
    graph_files : list
        list of paths to graph file
    class_labels_fname : str
        path to graph class label file
    
    Returns
    -------
    tuple
        kernel_row_x_id, kernel_row_y_id. The first is an list of integers each referencing a row in
        a kernel matrix and thereby a kernel vector corresponding to one of the graphs 
        in the dataset, the second is a list of class labels whose value is the classification
        of the graph in the same index of the first


    """
    graph_id_to_class_tuples = []
    graph_to_class_label_map = {l.split()[0]: int(l.split()[1].strip()) for l in open (class_labels_fname)}
    for graph_fname in graph_files:
        basename = os.path.basename(graph_fname)
        clabel = graph_to_class_label_map[basename]
        gidx = corpus._graph_name_to_id_map[graph_fname+extension]
        graph_id_to_class_tuples.append((gidx, clabel))

    graph_id_to_class_tuples.sort(key=lambda tup: tup[0])
    kernel_row_x_id, kernel_row_y_id = zip(*graph_id_to_class_tuples)

    return kernel_row_x_id, kernel_row_y_id


"""Data_reader module containing corpus construction utilities for PVDM models

"""

# Author: Paul Scherer

import numpy as np
import torch
from torch.utils.data import Dataset, DataLoader
from collections import defaultdict, Counter
from random import shuffle, randint, choice
from tqdm import tqdm

from .utils import get_files

np.random.seed(27)

#######################################################################################################
#######################################################################################################
# In memory version of pvdm datareader
#######################################################################################################
#######################################################################################################

class PVDMCorpus(Dataset):
	"""Class which representes all of the graph documents in a graph dataset 
	serves context for PVDM models, This version keeps the entire corpus 
	with negatives in memory which requires a larger initial creation 
	time but is much quicker at loading during training.

	Parameters
	----------
	corpus_dir : str
		path to folder with graph document files created in decomposition stage
	extension : str
		extension of the graph document files from which the corpus should be built
	max_files : int (default=0)
		the maximum number of files to include. Useful for debugging or other 
		artificial scenarios. The default of 0 includes all files with matching
		extension
	window_size : int (default=1)
		The number of context substructure patterns to be considered for every target. 
		This needs to be greater than 0.

	Returns
	-------
	self : PVDMCorpus 
		A corpus dataset that can be used with the PVDM with negative sampling model.

	"""

	NEGATIVE_TABLE_SIZE=1e8

	def __init__(self, corpus_dir=None, extension=".wld2", max_files=0, min_count=0, window_size=1):
		assert corpus_dir != None, "Please specify the path where the graph files are"
		assert window_size > 0, "Please make a window size that is greater than 0"
		self.corpus_dir = corpus_dir
		self.extension = extension
		self.graph_index = 0
		self.subgraph_index = 0
		self.epoch_flag = 0
		self.max_files = max_files
		self.graph_ids_for_batch_traversal = []
		self.min_count = min_count
		self.window_size = window_size

		self.negatives = []
		self.discards = []
		self.negpos = 0

		# Methods
		self.scan_and_load_corpus()
		self.initTableDiscards()
		self.initTableNegatives()
		self.pre_load_corpus()

	def scan_and_load_corpus(self):
		"""Gets the list of graph file paths, gives them number ids in a map and calls 
		scan_corpus also makes available a list of shuffled graph_ids
		"""

		print("#... Scanning and loading corpus from %s" % (self.corpus_dir))
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)
		self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)}
		self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}

		# Scan the corpus
		# This creates the alphabet and vocabulary of subgraphs
		subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

		self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
		shuffle(self.graph_ids_for_batch_traversal)

	def scan_corpus(self, min_count):
		"""Maps the graph files to a subgraph alphabet from which we create new_ids for the subgraphs
		which in turn get used by the skipgram architectures

		Parameters
		----------
		min_count : int
			The minimum number of times a subgraph pattern should appear across the graphs in 
			order to be considered part of the vocabulary.

		Returns
		-------
		(Optional) self._subgraph_to_id_map : dict
			dictionary of substructure pattern to int id map

		"""

		# Get all the subgraph names (the centers, ie without context around itself) ie first item of 
		# each line in the grapdoc files
		subgraphs = defaultdict(int)
		for fname in self.graph_fname_list:
			for l in open(fname).readlines():
				ss = l.split()[:self.window_size+1] # as context may have new subgraphs we havent seen				
				for s in ss:
					subgraphs[s] += 1


		# Frequency map of subgraph types
		subgraph_to_freq_map = subgraphs

		# Remove infrequent graph patterns if user specified a min_count
		if min_count:
			subgraph_to_freq_map = dict(subgraph_to_freq_map)
			for key in list(subgraph_to_freq_map.keys()):
				if subgraph_to_freq_map[key] < min_count:
					subgraph_to_freq_map.pop(key, None)

		# Also give each of the subgraph labels new int ids
		subgraph_to_id_map = {sg:i for i, sg, in enumerate(subgraph_to_freq_map.keys())}

		self._subgraph_to_freq_map = subgraph_to_freq_map
		self._subgraph_to_id_map = subgraph_to_id_map
		self._id_to_subgraph_map = {v:k  for k,v in subgraph_to_id_map.items()}
		self._subgraphcount = sum(subgraph_to_freq_map.values()) # token count, the number of subgraphs in all graphs


		self.num_graphs = len(self.graph_fname_list) # number of targets
		self.num_subgraphs = len(subgraph_to_id_map) # size of vocabulary

		# This is a list sorted by id, which contains the frequency of a subgraph appearing as a list
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])

		return self._subgraph_to_id_map

	def add_file(self, full_graph_path):
		"""This method is used to add new graphs into the corpus for 
		inductive learning of new unseen graphs

		Parameters
		----------
		full_graph_path : str
			path to graph document to be part of the new corpus

		Returns
		-------
		None
			New graph and its substructure patterns is made part of the corpus
		"""

		# Retrieve the graphs files and assign them internal ids for this method
		self.graph_fname_list = get_files(self.corpus_dir, self.extension, self.max_files)

		if full_graph_path in self.graph_fname_list:
			# if the graph is already in the corpus we may ignore it
			print("The graph %s is already in the corpus" % (full_graph_path))
			return
		else:
			self.graph_fname_list.append(full_graph_path) # add the new file
			self._graph_name_to_id_map = {g:i for i, g in enumerate(self.graph_fname_list)} # add to the end
			self._graph_name_to_id_map[full_graph_path] = self.num_graphs
			self._id_to_graph_name_map = {i:g for g, i in self._graph_name_to_id_map.items()}
				
			# Scan the corpus (ie explain in more detail in a sec)
			# This creates an "alphabet" and "vocabulary" of subgraphs
			subgraph_to_id_map = self.scan_corpus(min_count=self.min_count)

			self.graph_ids_for_batch_traversal = list(range(self.num_graphs))
			shuffle(self.graph_ids_for_batch_traversal)

			self.initTableDiscards()
			self.initTableNegatives()

	def initTableDiscards(self):
		self.subgraph_id_freq_map_as_list = [] # id of this list is the subgraph/word id and value is the frequency of the subgraph
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map_as_list.append(self._subgraph_to_freq_map[self._id_to_subgraph_map[i]])
		self.subgraph_id_freq_map = {}
		for i in range(len(self._subgraph_to_freq_map)):
			self.subgraph_id_freq_map[i] = self.subgraph_id_freq_map_as_list[i]

		t = 1e-4
		f = np.array(list(self.subgraph_id_freq_map.values())) / self._subgraphcount
		self.discards = np.sqrt(t/f)+(t/f)

	def initTableNegatives(self):
		pow_frequency = np.array(list(self.subgraph_id_freq_map.values()))**0.5
		words_pow = sum(pow_frequency)
		ratio = pow_frequency / words_pow
		count = np.round(ratio * PVDMCorpus.NEGATIVE_TABLE_SIZE)
		for sg_id, c in enumerate(count):
			self.negatives += [sg_id]*int(c)
		self.negatives = np.array(self.negatives)
		np.random.shuffle(self.negatives)

	def getNegatives(self, target, size):
		r"""Given target find a `size` number of negative samples by index

		Parameters
		----------
		target : int
			internal int id of the subgraph pattern
		size : int
			number of negative samples to find

		Returns
		-------
		response : [int]
			list of negative samples by internal int id

		"""

		response = self.negatives[self.negpos:self.negpos + size]
		self.negpos = (self.negpos + size) % len(self.negatives)
		while target in response: # check equality with target
			for i in np.where(response == target):
				response[i] = self.negatives[np.random.randint(0,len(self.negatives))]
		if len(response) != size:
			return np.concatenate((response, self.negatives[0:self.negpos]))
		return response


	def pre_load_corpus(self):
		"""Constructs and loads an entire context-pair dataset into memory
		
		"""

		print("#... Generating dataset in memory for quick dataloader access")
		self.context_pair_dataset = []

		for _ in tqdm(range(self._subgraphcount)):
			target_graph_ids = []
			target_subgraph_ids = []
			subgraph_contexts_ids = []

			# Extract a random graph and read its contents
			graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]] # pull out a random graph (its filename here)
			graph_contents = open(graph_name).readlines()
			
			# tldr: random graph traverser 
			# If we've looked at all the subgraphs we go to the next graph
			# We set the epoch flag true if we've also gone through n graphs (in a n graph dataset)
			# we then grab the next graph file whether that be the next graph in the shuffled list
			# or the first graph in a reshuffled list of graph files
			while self.subgraph_index >= len(graph_contents):
				self.subgraph_index = 0
				self.graph_index += 1
				if self.graph_index == len(self.graph_fname_list):
					self.graph_index = 0
					np.random.shuffle(self.graph_ids_for_batch_traversal)
					self.epoch_flag = True
				graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
				graph_contents = open(graph_name).readlines()


			# Read substructure patterns per graph
			line_id = self.subgraph_index
			# target_subgraph = graph_contents[line_id].split()[0] # first item on the line is the center
			# context_subgraphs = graph_contents[line_id].split()[1:self.window_size] # the contexts of the center we try to predict
			subgraphs_on_line = graph_contents[line_id].split()[:self.window_size+1]


			target_graph = graph_name

			# if target_subgraph in self._subgraph_to_id_map:
			# 	target_subgraph_ids.append(self._subgraph_to_id_map[target_subgraph]) # add the ids of the subgraph into the context
			# 	target_graph_ids.append(self._graph_name_to_id_map[target_graph]) # add the id of the graph for the target
			# 	temp_subgraph_contexts = []
			# 	for subgraph_context in context_subgraphs:
			# 		if subgraph_context in self._subgraph_to_id_map:
			# 			temp_subgraph_contexts.append(self._subgraph_to_id_map[subgraph_context])
			# 	subgraph_contexts_ids.append(temp_subgraph_contexts)


			# permuts = [target_subgraph] + context_subgraphs
			# for tgt, ctx in list(itertools.permutations(permuts, 2)):
			# 	if tgt in self._subgraph_to_id_map and ctx in self._subgraph_to_id_map:
			# 		target_graph_ids.append(self._graph_name_to_id_map[target_graph])
			# 		target_subgraph_ids.append(self._subgraph_to_id_map[tgt])
			# 		context_subgraph_ids.append(self._subgraph_to_id_map[ctx])

			# subgraphs_on_line = [target_subgraph] + context_subgraphs
			for i in range(len(subgraphs_on_line)):
				subgraphs_on_line_copy = subgraphs_on_line.copy()
				target_subgraph = subgraphs_on_line_copy.pop(i)
				if target_subgraph in self._subgraph_to_id_map:
					target_graph_ids.append(self._graph_name_to_id_map[target_graph])
					target_subgraph_ids.append(self._subgraph_to_id_map[target_subgraph])
					temp_subgraph_contexts = []
					for subgraph_context in subgraphs_on_line_copy:
						if subgraph_context in self._subgraph_to_id_map:
							temp_subgraph_contexts.append(self._subgraph_to_id_map[subgraph_context])
					while len(temp_subgraph_contexts) < self.window_size:
						subgraph_context = choice(subgraphs_on_line_copy)
						if subgraph_context in self._subgraph_to_id_map:
							temp_subgraph_contexts.append(self._subgraph_to_id_map[subgraph_context])
					subgraph_contexts_ids.append(temp_subgraph_contexts)

			# Dark lord hack which needs to be refactored later but also respects 
			# random uniform sampling of original
			ri = randint(0, len(target_graph_ids)-1)
			target_graph_ids = [target_graph_ids[ri]]
			target_subgraph_ids = [target_subgraph_ids[ri]]
			subgraph_contexts_ids = [subgraph_contexts_ids[ri]]



			# move on to the next subgraph
			self.subgraph_index += 1

			# if we've reached the end of the graph file (ie exhasuted all the subgraphs in it)
			# we move onto the next graph as above.
			while self.subgraph_index == len(graph_contents):
				self.subgraph_index = 0
				self.graph_index +=1 
				if self.graph_index == len(self.graph_fname_list):
					self.graph_index = 0
					np.random.shuffle(self.graph_ids_for_batch_traversal)
					self.epoch_flag = True

				graph_name = self.graph_fname_list[self.graph_ids_for_batch_traversal[self.graph_index]]
				graph_contents = open(graph_name).readlines()
			
			# Once we've built 'batch_size' number worth of targets and contexts
			# we zip them, shuffle them, and unzip the pairs in shuffled order (keeping the pairing)
			target_context_pairs = list(zip(target_graph_ids, target_subgraph_ids, subgraph_contexts_ids))
			shuffle(target_context_pairs)
			target_graph_ids, target_subgraph_ids, subgraph_contexts_ids= list(zip(*target_context_pairs))

			negatives_per_context = [self.getNegatives(x,10) for x in target_subgraph_ids]
			graph_targetSubgraph_subgraphContexts_negatives = [(graphTarget, subgraphTarget, subgraphContext, negatives) for (graphTarget, subgraphTarget, subgraphContext, negatives) in zip(target_graph_ids, target_subgraph_ids, subgraph_contexts_ids, negatives_per_context)]
			self.context_pair_dataset.append(graph_targetSubgraph_subgraphContexts_negatives)
		self.epoch_flag = False

	def __len__(self):
		"""Return the number of total number of subgraphs
		
		"""
		return len(self.context_pair_dataset) - 1

	def __getitem__(self, idx):
		"""Return the number of total number of subgraphs
		
		"""
		return self.context_pair_dataset[idx]

	@staticmethod
	def collate(batches):
		all_targets = [tgraph for batch in batches for tgraph, _, _, _ in batch if len(batch)>0]
		all_contexts = [tsubgraph for batch in batches for _, tsubgraph, _, _ in batch if len(batch)>0]
		all_subgraph_contexts = [csubgraph for batch in batches for _, _, csubgraph, _ in batch if len(batch)>0]
		all_neg_contexts = [neg_tsubgraph for batch in batches for _, _, _, neg_tsubgraph in batch if len(batch)>0]

		return torch.LongTensor(all_targets), torch.LongTensor(all_contexts), torch.LongTensor(all_subgraph_contexts), torch.LongTensor(all_neg_contexts)


if __name__ == '__main__':
	corpus_dir = "../data/dortmund_gexf/MUTAG"

	# In memory pvdm corpus
	in_mem_pvdm_corpus = PVDMCorpus(corpus_dir, extension=".awe_4_nodes", max_files=60, window_size=3)
	pvdm_dataloader = DataLoader(in_mem_pvdm_corpus, batch_size=4, shuffle=False, num_workers=0, collate_fn=in_mem_pvdm_corpus.collate)


"""
Module containining class definitions of trainers for skipgram models, 
which are partly used by Deep Graph Kernels

Author: Paul Scherer
"""

import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from tqdm import tqdm

# Internal
from .skipgram_data_reader import SkipgramCorpus, InMemorySkipgramCorpus
from .skipgram import Skipgram
from .utils import save_subgraph_embeddings

class Trainer(object):
	"""Handles corpus construction (hard drive version), skipgram initialization and training.

	Paramaters
	----------
	corpus_dir : str
		path to directory containing graph files
	extension : str
		extension used in graph documents produced after decomposition stage
	max_files : int
		the maximum number of graph files to consider, default of 0 uses all files
	output_fh : str
		the path to the file where embeddings should be saved
	emb_dimension : int (default=128)
		the desired dimension of the embeddings
	batch_size : int (default=32)
		the desired batch size
	epochs : int (default=100)
		the desired number of epochs for which the network should be trained
	initial_lr : float (default=1e-3)
		the initial learning rate
	min_count : int (default=1)
		the minimum number of times a pattern should occur across the dataset to 
		be considered part of the substructure pattern vocabulary

	Returns
	-------
	self : Trainer
		A Trainer instance

	"""

	def __init__(self, corpus_dir, extension, max_files, window_size, output_fh, emb_dimension=128, batch_size=32, epochs=100, initial_lr=1e-3, min_count=1):
		self.corpus = SkipgramCorpus(corpus_dir, extension, max_files, min_count, window_size)
		self.dataloader = DataLoader(self.corpus, batch_size, shuffle=False, num_workers=4, collate_fn = self.corpus.collate)

		self.corpus_dir = corpus_dir
		self.extension = extension
		self.max_files = max_files
		self.output_fh = output_fh
		self.emb_dimension = emb_dimension
		self.batch_size = batch_size
		self.epochs = epochs
		self.initial_lr = initial_lr
		self.min_count = min_count
		self.window_size = window_size

		self.num_targets = self.corpus.num_subgraphs # the special feature here is that we are learning subgraph reps
		self.vocab_size = self.corpus.num_subgraphs

		self.skipgram = Skipgram(self.num_targets, self.vocab_size, self.emb_dimension)

		if torch.cuda.is_available():
			self.device = torch.device("cuda")
			self.skipgram.cuda()
		else:
			self.device = torch.device("cpu")

	def train(self):
		"""Train the network with the settings used to initialise the Trainer"""
		for epoch in range(self.epochs):
			print("### Epoch: " + str(epoch))
			optimizer = optim.Adagrad(self.skipgram.parameters(), lr=self.initial_lr)
			scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, len(self.dataloader))

			running_loss = 0.0
			for sample_batched in tqdm(self.dataloader):

				if len(sample_batched[0]) > 1:
					pos_target = sample_batched[0].to(self.device)
					pos_context = sample_batched[1].to(self.device)
					neg_context = sample_batched[2].to(self.device)

					optimizer.zero_grad()
					loss = self.skipgram.forward(pos_target, pos_context, neg_context) # the loss is integrated into the forward function
					loss.backward()
					optimizer.step()
					scheduler.step()

					running_loss = running_loss * 0.9 + loss.item() * 0.1
			print(" Loss: " + str(running_loss))

		final_embeddings = self.skipgram.target_embeddings.weight.cpu().data.numpy()
		save_subgraph_embeddings(self.corpus, final_embeddings, self.output_fh)
		return(final_embeddings)


class InMemoryTrainer(object):
	"""Handles corpus construction (in-memory version), PVDBOW initialization and training.

	Paramaters
	----------
	corpus_dir : str
		path to directory containing graph files
	extension : str
		extension used in graph documents produced after decomposition stage
	max_files : int
		the maximum number of graph files to consider, default of 0 uses all files
	output_fh : str
		the path to the file where embeddings should be saved
	emb_dimension : int (default=128)
		the desired dimension of the embeddings
	batch_size : int (default=32)
		the desired batch size
	epochs : int (default=100)
		the desired number of epochs for which the network should be trained
	initial_lr : float (default=1e-3)
		the initial learning rate
	min_count : int (default=1)
		the minimum number of times a pattern should occur across the dataset to 
		be considered part of the substructure pattern vocabulary

	Returns
	-------
	self : InMemoryTrainer
		A trainer instance which has the dataset stored in memory for fast access
	"""

	def __init__(self, corpus_dir, extension, max_files, window_size, output_fh, emb_dimension=128, batch_size=32, epochs=100, initial_lr=1e-3, min_count=1):
		self.corpus = InMemorySkipgramCorpus(corpus_dir, extension, max_files, min_count, window_size)
		self.dataloader = DataLoader(self.corpus, batch_size, shuffle=False, num_workers=4, pin_memory=True, collate_fn = self.corpus.collate)

		self.corpus_dir = corpus_dir
		self.extension = extension
		self.max_files = max_files
		self.output_fh = output_fh
		self.emb_dimension = emb_dimension
		self.batch_size = batch_size
		self.epochs = epochs
		self.initial_lr = initial_lr
		self.min_count = min_count
		self.window_size = window_size

		self.num_targets = self.corpus.num_subgraphs # the special feature here is that we are learning subgraph reps
		self.vocab_size = self.corpus.num_subgraphs

		self.skipgram = Skipgram(self.num_targets, self.vocab_size, self.emb_dimension)

		if torch.cuda.is_available():
			self.device = torch.device("cuda")
			self.skipgram.cuda()
		else:
			self.device = torch.device("cpu")

	def train(self):
		"""Train the network with the settings used to initialise the Trainer"""
		for epoch in range(self.epochs):
			print("### Epoch: " + str(epoch))
			optimizer = optim.Adagrad(self.skipgram.parameters(), lr=self.initial_lr)
			scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, len(self.dataloader))

			running_loss = 0.0
			for sample_batched in tqdm(self.dataloader):

				if len(sample_batched[0]) > 1:
					pos_target = sample_batched[0].to(self.device)
					pos_context = sample_batched[1].to(self.device)
					neg_context = sample_batched[2].to(self.device)

					optimizer.zero_grad()
					loss = self.skipgram.forward(pos_target, pos_context, neg_context) # the loss is integrated into the forward function
					loss.backward()
					optimizer.step()
					scheduler.step()

					running_loss = running_loss * 0.9 + loss.item() * 0.1
			print(" Loss: " + str(running_loss))

		final_embeddings = self.skipgram.target_embeddings.weight.cpu().data.numpy()
		save_subgraph_embeddings(self.corpus, final_embeddings, self.output_fh)
		return(final_embeddings)


if __name__ == '__main__':
	pass
		

